#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
import copy
import glob
import os
from argparse import RawTextHelpFormatter

import rapidtide.externaltools as tide_exttools
import rapidtide.filter as tide_filt
import rapidtide.io as tide_io

""


def _get_parser():
    class FullPaths(argparse.Action):
        """Expand user- and relative-paths"""

        def __call__(self, parser, namespace, values, option_string=None):
            if values == "":
                setattr(namespace, self.dest, "__EMPTY__")
            else:
                setattr(namespace, self.dest, os.path.abspath(os.path.expanduser(values)))

    def is_file(filename):
        """Checks if a file exists"""
        if not os.path.exists(filename):
            msg = "{0} does not exist".format(filename)
            raise argparse.ArgumentTypeError(msg)
        else:
            return filename

    parser = argparse.ArgumentParser(
        description=(
            "Align an anatomic image with the preprocessed T1 in an fmriprep derivatives directory.\n"
            "fmriprep does not align other anatomic images to the T1 by default.  This fixes that.\n"
        ),
        formatter_class=RawTextHelpFormatter,
        allow_abbrev=False,
    )

    parser.add_argument(
        "sourceimage",
        help="cine output file from happy, in tide_io.nifti format",
        action=FullPaths,
        type=is_file,
    )

    parser.add_argument(
        "--scalefac",
        help="scale factor to exaggerate motion",
        type=float,
        default=20.0,
    )

    parser.add_argument(
        "--threads",
        help="number of threads to use for registration",
        type=int,
        default=1,
    )

    parser.add_argument(
        "--finaltarget",
        help="align all output images to this image in the final step",
        action=FullPaths,
        type=is_file,
        default=None,
    )

    parser.add_argument(
        "--splinesize",
        help="spline size in voxels",
        type=int,
        default=10,
    )

    parser.add_argument(
        "--transformtype",
        help="Transform type.  Options are\n"
        "\tr: rigid\n"
        "\ta: rigid + affine\n"
        "\ts: rigid + affine + deformable syn\n"
        "\tb: rigid + affine + deformable b-spline syn\n",
        default="b",
    )

    parser.add_argument(
        "--rescaleonly",
        help="only perform the warp scaling (assumes a previous full run).",
        action="store_true",
    )

    parser.add_argument(
        "--redofinalalignment",
        help="recalculate the transformation between epi and anatomic",
        action="store_true",
    )

    parser.add_argument(
        "--fake",
        help="Print what operations will be performed, but don't do them.",
        action="store_true",
    )

    parser.add_argument(
        "--debug",
        help="output additional debugging information",
        action="store_true",
    )

    return parser


def filterintime(filelist, extractedfile, workdir, filtcycles, dumpinput=False):
    numpoints = len(filelist)
    filelist = filelist * filtcycles
    filterinput = os.path.join(workdir, "filterinput")
    filterdata, filter_hdr = tide_io.niftimerge(
        filelist, filterinput, returndata=True, writetodisk=dumpinput
    )
    outdata = np.zeros_like(filterdata)
    thedims = filterdata.shape
    print(thedims)
    numvoxels = thedims[0] * thedims[1] * thedims[2]
    thefilter = tide_filt.NoncausalFilter(filtertype="arb")
    hcutoff = 1.0 / 2.0
    lcutoff = 1.0 / 32.0
    thefilter.setfreqs(lcutoff, lcutoff, hcutoff, hcutoff)

    filterdata_byvoxel = filterdata.reshape((numvoxels, thedims[3], thedims[4]))
    outdata_byvoxel = outdata.reshape((numvoxels, thedims[3], thedims[4]))
    for direction in range(thedims[4]):
        print("filtering direction", direction)
        for voxel in range(numvoxels):
            filteredtc = thefilter.apply(1.0, filterdata_byvoxel[voxel, :, direction])
            outdata_byvoxel[voxel, :, direction] = filteredtc + 0.0

    tide_io.savetonifti(outdata[:, :, :, :, :], filter_hdr, "filteroutdata")

    theheader = copy.deepcopy(filter_hdr)
    theheader["dim"][4] = numpoints
    startpt = int((filtcycles // 2) * numpoints)
    endpt = startpt + numpoints
    tide_io.savetonifti(outdata[:, :, :, startpt:endpt, :], theheader, extractedfile)


def dohappywarp(
    sourceimage,
    transformtype="b",
    scalefac=20.0,
    threads=1,
    splinesize=10,
    rescaleonly=False,
    fake=False,
    finaltarget=None,
    redofinalalignment=False,
    debug=False,
):
    if debug:
        print("arguments to dohappywarp:")
        print("\tsourceimage:", sourceimage)
        print("\ttransformtype:", transformtype)
        print("\tscalefac:", scalefac)
        print("\tthreads:", threads)
        print("\tsplinesize:", splinesize)
        print("\trescaleonly:", rescaleonly)
        print("\tfinaltarget:", finaltarget)
        print("\tredofinalalignment:", redofinalalignment)
        print("\tfake:", fake)

    refimage = "cinemean.nii.gz"
    filtcycles = 5
    sourcedir, sourcefile = os.path.split(sourceimage)
    workdir = os.path.join(sourcedir, "warptemp")
    try:
        os.makedirs(workdir)
    except FileExistsError:
        pass
    reffullname = os.path.join(workdir, refimage)
    if debug:
        print("reffullname:", reffullname)

    # make a mean image
    fslmathscmd = ["fslmaths"]
    fslmathscmd += [sourceimage]
    fslmathscmd += ["-Tmean"]
    fslmathscmd += [reffullname]
    tide_exttools.runcmd(fslmathscmd, fake=fake)

    # intensity correct the mean image
    tide_exttools.n4correct(reffullname, workdir, fake=fake)
    correctedreffullname = reffullname.replace(".nii.gz", "_n4.nii.gz")

    # if there is a target image, align the mean image with it
    if finaltarget is not None and (not rescaleonly or redofinalalignment):
        tide_exttools.antsalign(
            finaltarget,
            correctedreffullname,
            os.path.join(workdir, "tofinal_"),
            transformtype="b",
            threads=threads,
            fake=fake,
        )

    if not rescaleonly:
        # split the cine file
        splitroot = os.path.join(workdir, "cineframe_")
        tide_io.niftisplit(sourceimage, splitroot)

        # intensity correct the split files
        filespec = splitroot + "[0123456789][0123456789][0123456789][0123456789].nii.gz"
        print(filespec)
        framefiles = glob.glob(filespec)
        print("framefiles:", framefiles)
        framefiles.sort()
        for framefile in framefiles:
            tide_exttools.n4correct(framefile, workdir, fake=fake)

        # make a sorted list of the intensity corrected cine frames
        correctedframefiles = glob.glob(splitroot + "*_n4.nii.gz")
        correctedframefiles.sort()
        numtimepoints = len(correctedframefiles)

        # find the alignment of every the mean image to every frame
        doinitalign = True
        alignroot = os.path.join(workdir, "alignedtomean_")
        if doinitalign:
            for correctedframefile in correctedframefiles:
                thenumber = correctedframefile[len(splitroot) : len(splitroot) + 4]
                theoutput = correctedframefile.replace(splitroot, alignroot).replace(
                    ".nii.gz", "_"
                )
                if debug:
                    print("thenumber:", thenumber)
                    print("initfile:", correctedframefile)
                    print("theoutput:", theoutput)
                tide_exttools.antsalign(
                    correctedframefile,
                    correctedreffullname,
                    theoutput,
                    transformtype=transformtype,
                    threads=threads,
                    splinesize=splinesize,
                    fake=fake,
                )

    # lowpass filter the warp data in time
    alignroot = os.path.join(workdir, "alignedtomean_")
    extractedfile = os.path.join(workdir, "filteredWarp")
    filelist = glob.glob(alignroot + "[0123456789]*_1Warp.nii.gz")
    filelist.sort()
    filterintime(filelist, extractedfile, workdir, filtcycles, dumpinput=True)

    # scale the warp file
    extractedscaledfile = os.path.join(workdir, "filteredWarp_scaled.nii.gz")
    scalecmd = []
    scalecmd += ["fslmaths"]
    scalecmd += [extractedfile]
    scalecmd += ["-mul", str(scalefac)]
    scalecmd += [extractedscaledfile]
    tide_exttools.runcmd(scalecmd, fake=fake)

    # split the filtered warps into timepoints
    splitunscaledwarproot = os.path.join(workdir, "filtered_Warp_")
    splitscaledwarproot = os.path.join(workdir, "filtered_Warp_scaled_")
    tide_io.niftisplit(extractedfile, splitunscaledwarproot)
    tide_io.niftisplit(extractedscaledfile, splitscaledwarproot)

    # now apply the warp files to the mean
    splitroot = os.path.join(workdir, "filtered_Warp_")
    alignedroot = os.path.join(workdir, "filtered_Warped_")

    warpfiles = glob.glob(splitroot + "[0123456789]*.nii.gz")
    warpfiles.sort()
    targetname = correctedreffullname

    for warpfile in warpfiles:
        theoutput = warpfile.replace(splitroot, alignedroot)
        translist = [warpfile]
        if finaltarget is not None:
            translist = [
                os.path.join(workdir, "tofinal_1Warp.nii.gz"),
                os.path.join(workdir, "tofinal_0GenericAffine.mat"),
            ] + translist
        tide_exttools.antsapply(targetname, targetname, theoutput, translist, fake=fake)
        tide_exttools.antsapply(
            targetname,
            targetname,
            theoutput.replace("Warped_", "Warped_scaled_"),
            [warpfile.replace("Warp_", "Warp_scaled_")],
            fake=fake,
        )

    # merge the warped files
    print("out of transform loop")
    mergedoutput = os.path.join(sourcedir, "warpedcine")
    mergedoutputscaled = os.path.join(sourcedir, "warpedcinescaled")
    filespec = alignedroot + "[0123456789]*.nii.gz"
    print(filespec)
    filelist = glob.glob(filespec)
    filelist.sort()
    tide_io.niftimerge(filelist, mergedoutput)
    filespec = alignedroot + "scaled_*.nii.gz"
    print(filespec)
    filelist = glob.glob(filespec)
    filelist.sort()
    tide_io.niftimerge(filelist, mergedoutputscaled)


def entrypoint():
    # get the command line parameters
    try:
        args = _get_parser().parse_args()
    except SystemExit:
        _get_parser().print_help()
        raise

    if args.debug:
        print(args)

    dohappywarp(
        args.sourceimage,
        transformtype=args.transformtype,
        splinesize=args.splinesize,
        scalefac=args.scalefac,
        threads=args.threads,
        rescaleonly=args.rescaleonly,
        fake=args.fake,
        finaltarget=args.finaltarget,
        redofinalalignment=args.redofinalalignment,
        debug=args.debug,
    )


if __name__ == "__main__":
    entrypoint()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np

import rapidtide.util as tide_util


def test_numpy2shared(debug=False):
    vectorlen = 1000
    for intype in [np.float32, np.float64]:
        sourcevector = np.random.normal(size=vectorlen).astype(intype)
        if debug:
            print(f"{intype=}, {sourcevector.size=}, {sourcevector.dtype=}")
        for outtype in [np.float32, np.float64]:

            destvector, shm = tide_util.numpy2shared(sourcevector, outtype)
            if debug:
                print(f"\t{outtype=}, {destvector.size=}, {destvector.dtype=}")

            # check everything
            assert destvector.dtype == outtype
            assert destvector.size == sourcevector.size
            np.testing.assert_almost_equal(sourcevector, destvector, 3)

            # clean up
            tide_util.cleanup_shm(shm)


def test_allocshared(debug=False):
    datashape = (10, 10, 10)
    for outtype in [np.float32, np.float64]:
        destarray, shm = tide_util.allocshared(datashape, outtype)
        if debug:
            print(f"{outtype=}, {destarray.size=}, {destarray.dtype=}")

        # check everything
        assert destarray.dtype == outtype
        assert destarray.size == np.prod(datashape)

        # clean up if needed
        tide_util.cleanup_shm(shm)


if __name__ == "__main__":
    test_numpy2shared(debug=True)
    test_allocshared(debug=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
import platform
import sys
import time
from argparse import Namespace
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray

import rapidtide.filter as tide_filt
import rapidtide.io as tide_io
import rapidtide.miscmath as tide_math
import rapidtide.util as tide_util
import rapidtide.workflows.parser_funcs as pf


def _get_parser() -> Any:
    """
    Construct and return an argument parser for the gmscalc tool.

    This function sets up an `argparse.ArgumentParser` with required and optional
    arguments needed to run the global mean signal calculation and filtering
    pipeline. It includes support for specifying input data, output root, data mask,
    normalization options, smoothing, and debugging.

    Returns
    -------
    argparse.ArgumentParser
        Configured argument parser with all required and optional arguments
        for the gmscalc tool.

    Notes
    -----
    The parser is configured with `allow_abbrev=False` to enforce full argument
    names and avoid ambiguity.

    Examples
    --------
    >>> parser = _get_parser()
    >>> args = parser.parse_args(['data.nii', 'output_root'])
    >>> print(args.datafile)
    'data.nii'
    """
    parser = argparse.ArgumentParser(
        prog="gmscalc",
        description="Calculate the global mean signal, and filtered versions",
        allow_abbrev=False,
    )

    # Required arguments
    parser.add_argument(
        "datafile",
        type=str,
        help=(
            "The name of a 4 dimensional nifti files (3 spatial dimensions and a "
            "subject dimension)."
        ),
    )
    parser.add_argument("outputroot", type=str, help="The root name for the output nifti files.")

    # Optional arguments
    parser.add_argument(
        "--dmask",
        dest="datamaskname",
        type=lambda x: pf.is_valid_file(parser, x),
        action="store",
        metavar="DATAMASK",
        help=("Use DATAMASK to specify which voxels in the data to use."),
        default=None,
    )
    pf.addnormalizationopts(parser, normtarget="timecourses", defaultmethod="None")

    parser.add_argument(
        "--normfirst",
        dest="normfirst",
        action="store_true",
        help=("Normalize before filtering, rather than after."),
        default=False,
    )
    parser.add_argument(
        "--smooth",
        dest="sigma",
        type=lambda x: pf.is_float(parser, x),
        action="store",
        metavar="SIGMA",
        help=("Spatially smooth the input data with a SIGMA mm kernel prior to calculation."),
        default=0.0,
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help=("Print out additional debugging information."),
        default=False,
    )
    return parser


def makecommandlinelist(
    arglist: Any, starttime: Any, endtime: Any, extra: Optional[Any] = None
) -> None:
    """
    Create a list of command line information for logging purposes.

    This function generates a list of descriptive strings containing processing
    information including date, duration, version details, and the actual command
    that was executed.

    Parameters
    ----------
    arglist : Any
        List of command line arguments to be joined into a command string.
    starttime : Any
        Start time of the process, typically a timestamp.
    endtime : Any
        End time of the process, typically a timestamp.
    extra : Any, optional
        Additional descriptive text to include in the output list. Default is None.

    Returns
    -------
    list of str
        List containing the following elements in order:
        - Processing date and time
        - Processing duration
        - Node and version information
        - Extra information (if provided)
        - The actual command line string

    Notes
    -----
    The function uses `time.strftime` to format the start time and `tide_util.version()`
    to retrieve version information. The command line is constructed by joining
    the `arglist` elements with spaces.

    Examples
    --------
    >>> import time
    >>> args = ['python', 'script.py', '--input', 'data.txt']
    >>> start = time.time()
    >>> # ... some processing ...
    >>> end = time.time()
    >>> info = makecommandlinelist(args, start, end)
    >>> print(info[0])
    '# Processed on Mon, 01 Jan 2024 12:00:00 UTC.'
    """
    # get the processing date
    dateline = (
        "# Processed on "
        + time.strftime("%a, %d %b %Y %H:%M:%S %Z", time.localtime(starttime))
        + "."
    )
    timeline = f"# Processing took {endtime - starttime:.3f} seconds."

    # diagnostic information about version
    (
        release_version,
        dummy,
        git_date,
        dummy,
    ) = tide_util.version()

    nodeline = "# " + " ".join(
        [
            "Using",
            platform.node(),
            "(",
            release_version + ",",
            git_date,
            ")",
        ]
    )

    # and the actual command issued
    commandline = " ".join(arglist)

    if extra is not None:
        return [dateline, timeline, nodeline, "# " + extra, commandline]
    else:
        return [dateline, timeline, nodeline, commandline]


def gmscalc_main() -> None:
    """
    Main function to calculate global mean signal (GMS) from fMRI data.

    This function reads NIfTI-formatted fMRI data, applies optional smoothing,
    masks the data if a mask is provided, and computes the global mean signal
    across valid voxels. It then applies low-frequency (LFO) and high-frequency
    (HF) filtering to the global signal and writes the results to text files.

    The function uses the `tide_io` module for reading and writing data, and
    `tide_filt` and `tide_math` for filtering and normalization.

    Notes
    -----
    The function expects a command-line interface to be set up with `_get_parser()`
    and uses `sys.argv` to parse arguments. It prints diagnostic information
    during execution.

    Examples
    --------
    Assuming the script is called as `gmscalc_main.py` and properly configured:

    >>> gmscalc_main()

    This will read the input data, perform processing, and write output files
    with names based on the `outputroot` argument.

    See Also
    --------
    tide_io.readfromnifti : Reads NIfTI files.
    tide_io.writevec : Writes vectors to text files.
    tide_filt.ssmooth : Applies spatial smoothing.
    tide_filt.NoncausalFilter : Applies non-causal filtering.
    tide_math.normalize : Normalizes a signal.
    """
    try:
        args = _get_parser().parse_args()
    except SystemExit:
        _get_parser().print_help()
        raise

    runstarttime = time.time()

    # now read in the data
    print("reading datafile")
    (
        datafile_img,
        datafile_data,
        datafile_hdr,
        thedims,
        thesizes,
    ) = tide_io.readfromnifti(args.datafile)

    xsize, ysize, numslices, timepoints = tide_io.parseniftidims(thedims)
    xdim, ydim, slicethickness, tr = tide_io.parseniftisizes(thesizes)
    print(f"Datafile shape is {datafile_data.shape}")

    # smooth the data
    if args.sigma > 0.0:
        print("smoothing data")
        for i in range(timepoints):
            datafile_data[:, :, :, i] = tide_filt.ssmooth(
                xdim, ydim, slicethickness, args.sigma, datafile_data[:, :, :, i]
            )
        print("done smoothing data")

    if args.datamaskname is not None:
        print("reading in mask array")
        (
            datamask_img,
            datamask_data,
            datamask_hdr,
            datamaskdims,
            datamasksizes,
        ) = tide_io.readfromnifti(args.datamaskname)
        if not tide_io.checkspacematch(datafile_hdr, datamask_hdr):
            print("Data and mask dimensions do not match - exiting.")
            sys.exit()
        print("done reading in mask array")
    else:
        datamask_data = np.ones_like(datafile_data[:, :, :, 0])

    # now reformat from x, y, z, time to voxelnumber, measurement, subject
    numvoxels = int(xsize) * int(ysize) * int(numslices)
    mask_in_vox = datamask_data.reshape((numvoxels))

    print("finding valid voxels")
    validvoxels = np.where(mask_in_vox > 0)[0]
    numvalid = int(len(validvoxels))

    data_in_voxacq = datafile_data.reshape((numvoxels, timepoints))
    valid_in_voxacq = data_in_voxacq[validvoxels, :]

    # calculate each voxel's mean over time
    mean_in_voxacq = np.mean(valid_in_voxacq, axis=1)

    # calculate mean timecourse
    gms = np.mean(valid_in_voxacq, axis=0)

    # now filter
    lfofilter = tide_filt.NoncausalFilter("lfo")
    hffilter = tide_filt.NoncausalFilter("arb")
    lowerpass = 0.175
    lowerstop = 0.95 * lowerpass
    hffilter.setfreqs(lowerstop, lowerpass, 10.0, 10.0)
    gms_lfo = tide_math.normalize(lfofilter.apply(1.0 / tr, gms), method=args.normmethod)
    gms_hf = tide_math.normalize(hffilter.apply(1.0 / tr, gms), method=args.normmethod)

    tide_io.writevec(gms, args.outputroot + "_gms.txt")
    tide_io.writevec(gms_lfo, args.outputroot + "_gmslfo.txt")
    tide_io.writevec(gms_hf, args.outputroot + "_gmshf.txt")
    runendtime = time.time()
    thecommandfilelines = makecommandlinelist(sys.argv, runstarttime, runendtime)
    tide_io.writevec(thecommandfilelines, args.outputroot + "_commandline.txt")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
import os

import numpy as np

import rapidtide.multiproc as tide_multiproc
import rapidtide.refinedelay as tide_refinedelay
import rapidtide.util as tide_util
import rapidtide.workflows.parser_funcs as pf


def _get_parser():
    """
    Argument parser for stupidramtricks
    """
    parser = argparse.ArgumentParser(
        prog="stupidramtricks",
        description="Make shared memory blocks, do something with them.",
        allow_abbrev=False,
    )

    # Required arguments
    parser.add_argument(
        "--numvoxels",
        dest="numvoxels",
        action="store",
        type=int,
        metavar="NUMVOXELS",
        help=("Set blocks to have NUMVOXELS voxels."),
        default=230000,
    )
    # Required arguments
    parser.add_argument(
        "--numtimepoints",
        dest="numtimepoints",
        action="store",
        type=int,
        metavar="NUMPOINTS",
        help=("Set blocks to have NUMPOINTS timespoints."),
        default=500,
    )

    parser.add_argument(
        "--nprocs",
        dest="nprocs",
        action="store",
        type=int,
        metavar="NPROCS",
        help=(
            "Use NPROCS worker processes for multiprocessing. "
            "Setting NPROCS to less than 1 sets the number of "
            "worker processes to n_cpus."
        ),
        default=1,
    )
    parser.add_argument(
        "--usesharedmem",
        dest="usesharedmem",
        action="store_true",
        help=("Use shared memory"),
        default=False,
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help=("Print debug messages"),
        default=False,
    )

    return parser


def stupidramtricks(args):
    # get the pid of the parent process
    args.pid = os.getpid()

    args.internalprecision = "single"
    args.outputprecision = "double"

    print("Start")

    if args.nprocs < 1:
        args.nprocs = tide_multiproc.maxcpus()

    if args.internalprecision == "double":
        rt_floattype = np.float64
    else:
        rt_floattype = np.float32

    # set the output precision
    if args.outputprecision == "double":
        rt_outfloattype = np.float64
    else:
        rt_outfloattype = np.float32

    # select the voxels in the mask
    print("setting sizes")
    numvalidspatiallocs = args.numvoxels
    if args.debug:
        print(f"{numvalidspatiallocs=}")
    internalvalidspaceshape = numvalidspatiallocs
    derivaxissize = 2
    internalvalidspaceshapederivs = (
        internalvalidspaceshape,
        derivaxissize,
    )
    internalvalidfmrishape = (numvalidspatiallocs, args.numtimepoints)
    if args.debug:
        print(f"validvoxels shape = {numvalidspatiallocs}")
        print(f"internalvalidfmrishape shape = {internalvalidfmrishape}")

    fmridata = np.zeros(internalvalidfmrishape, dtype=rt_outfloattype)

    if args.usesharedmem:
        if args.debug:
            print("allocating shared memory")
        # first move fmridata into shared memory
        fmridata, fmridata_shm = tide_util.numpy2shared(
            fmridata, rt_floattype, name=f"fmridata_{args.pid}"
        )
        sLFOfitmean, sLFOfitmean_shm = tide_util.allocshared(
            internalvalidspaceshape, rt_outfloattype
        )
        rvalue, rvalue_shm = tide_util.allocshared(internalvalidspaceshape, rt_outfloattype)
        r2value, r2value_shm = tide_util.allocshared(internalvalidspaceshape, rt_outfloattype)
        fitNorm, fitNorm_shm = tide_util.allocshared(
            internalvalidspaceshapederivs, rt_outfloattype
        )
        fitcoeff, fitcoeff_shm = tide_util.allocshared(
            internalvalidspaceshapederivs, rt_outfloattype
        )
        movingsignal, movingsignal_shm = tide_util.allocshared(
            internalvalidfmrishape, rt_outfloattype
        )
        lagtc, lagtc_shm = tide_util.allocshared(internalvalidfmrishape, rt_floattype)
        filtereddata, filtereddata_shm = tide_util.allocshared(
            internalvalidfmrishape, rt_outfloattype
        )

        location = "in shared memory"
    else:
        if args.debug:
            print("allocating memory")
        sLFOfitmean = np.zeros(internalvalidspaceshape, dtype=rt_outfloattype)
        rvalue = np.zeros(internalvalidspaceshape, dtype=rt_outfloattype)
        r2value = np.zeros(internalvalidspaceshape, dtype=rt_outfloattype)
        fitNorm = np.zeros(internalvalidspaceshapederivs, dtype=rt_outfloattype)
        fitcoeff = np.zeros(internalvalidspaceshapederivs, dtype=rt_outfloattype)
        movingsignal = np.zeros(internalvalidfmrishape, dtype=rt_outfloattype)
        lagtc = np.zeros(internalvalidfmrishape, dtype=rt_floattype)
        filtereddata = np.zeros(internalvalidfmrishape, dtype=rt_outfloattype)
        location = "locally"

    totalbytes = (
        fmridata.nbytes
        + sLFOfitmean.nbytes
        + rvalue.nbytes
        + r2value.nbytes
        + fitNorm.nbytes
        + fitcoeff.nbytes
        + movingsignal.nbytes
        + lagtc.nbytes
        + filtereddata.nbytes
    )

    thesize, theunit = tide_util.format_bytes(totalbytes)
    print(f"allocated {thesize:.3f} {theunit} {location}")

    """regressderivratios, regressrvalues = tide_refinedelay.getderivratios(
        fmri_data_valid,
        validvoxels,
        initial_fmri_x,
        lagtimes_valid,
        corrmask_valid,
        genlagtc,
        mode,
        outputname,
        oversamptr,
        sLFOfitmean,
        rvalue,
        r2value,
        fitNorm[:, :2],
        fitcoeff[:, :2],
        movingsignal,
        lagtc,
        filtereddata,
        LGR,
        TimingLGR,
        therunoptions,
        debug=args.debug,
    )"""

    # clean up shared memory
    if args.usesharedmem:
        tide_util.cleanup_shm(fmridata_shm)
        tide_util.cleanup_shm(sLFOfitmean_shm)
        tide_util.cleanup_shm(rvalue_shm)
        tide_util.cleanup_shm(r2value_shm)
        tide_util.cleanup_shm(fitNorm_shm)
        tide_util.cleanup_shm(fitcoeff_shm)
        tide_util.cleanup_shm(movingsignal_shm)
        tide_util.cleanup_shm(lagtc_shm)
        tide_util.cleanup_shm(filtereddata_shm)


def process_args(inputargs=None):
    """
    Compile arguments for stupidramtrics workflow.
    """
    args, argstowrite = pf.setargs(_get_parser, inputargs=inputargs)
    return args


def entrypoint():
    pf.generic_init(_get_parser, stupidramtricks)


if __name__ == "__main__":
    entrypoint()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

import rapidtide.miscmath as tide_math
import rapidtide.resample as tide_resample
from rapidtide.correlate import AliasedCorrelator


def aliasedcorrelate(
    hiressignal,
    hires_Fs,
    lowressignal,
    lowres_Fs,
    timerange,
    hiresstarttime=0.0,
    lowresstarttime=0.0,
    padtime=30.0,
):
    """Perform an aliased correlation.

    This function is deprecated, and is retained here as a reference against
    which to test AliasedCorrelator.

    Parameters
    ----------
    hiressignal: 1D array
        The unaliased waveform to match
    hires_Fs: float
        The sample rate of the unaliased waveform
    lowressignal: 1D array
        The aliased waveform to match
    lowres_Fs: float
        The sample rate of the aliased waveform
    timerange: 1D array
        The delays for which to calculate the correlation function

    Returns
    -------
    corrfunc: 1D array
        The correlation function evaluated at timepoints of timerange
    """
    highresaxis = np.arange(0.0, len(hiressignal)) * (1.0 / hires_Fs) - hiresstarttime
    lowresaxis = np.arange(0.0, len(lowressignal)) * (1.0 / lowres_Fs) - lowresstarttime
    tcgenerator = tide_resample.FastResampler(highresaxis, hiressignal, padtime=padtime)
    targetsignal = tide_math.corrnormalize(lowressignal)
    corrfunc = np.zeros_like(timerange)
    for i in range(len(timerange)):
        aliasedhiressignal = tide_math.corrnormalize(tcgenerator.yfromx(lowresaxis + timerange[i]))
        corrfunc[i] = np.dot(aliasedhiressignal, targetsignal)
    return corrfunc


def test_aliasedcorrelate(displayplots=False):
    Fs_hi = 10.0
    Fs_lo = 1.0
    numsteps = 10
    siginfo = [[1.0, 1.36129345], [0.33, 2.0]]
    modamp = 0.01
    inlenhi = 1000
    inlenlo = 100
    offset = 0.5
    width = 2.5
    rangepts = 101
    timerange = np.linspace(0.0, width, num=101) - width / 2.0
    hiaxis = np.linspace(0.0, 2.0 * np.pi * inlenhi / Fs_hi, num=inlenhi, endpoint=False)
    loaxis = np.linspace(0.0, 2.0 * np.pi * inlenlo / Fs_lo, num=inlenlo, endpoint=False)
    sighi = np.zeros_like(hiaxis)
    siglo = np.zeros_like(loaxis)
    for theinfo in siginfo:
        sighi += theinfo[0] * np.sin(theinfo[1] * hiaxis)
        siglo += theinfo[0] * np.sin(theinfo[1] * loaxis)
    aliasedcorrelate_result = aliasedcorrelate(
        sighi, Fs_hi, siglo, Fs_lo, timerange, padtime=width
    )

    theAliasedCorrelator = AliasedCorrelator(sighi, Fs_hi, numsteps)
    aliasedcorrelate_result2 = theAliasedCorrelator.apply(siglo, 0)

    if displayplots:
        plt.figure()
        # plt.ylim([-1.0, 3.0])
        plt.plot(hiaxis, sighi, "k")
        plt.scatter(loaxis, siglo, c="r")
        plt.legend(["sighi", "siglo"])

        plt.figure()
        plt.plot(timerange, aliasedcorrelate_result, "k")
        plt.plot(timerange, aliasedcorrelate_result2, "r")
        print("maximum occurs at offset", timerange[np.argmax(aliasedcorrelate_result)])

        plt.show()

    # assert (fastcorrelate_result == stdcorrelate_result).all
    aethresh = 10
    # np.testing.assert_almost_equal(fastcorrelate_result, stdcorrelate_result, aethresh)


if __name__ == "__main__":
    mpl.use("TkAgg")
    test_aliasedcorrelate(displayplots=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
import sys
from argparse import Namespace
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray
from scipy.stats import skew

import rapidtide.io as tide_io
import rapidtide.workflows.parser_funcs as pf


def _get_parser() -> Any:
    """
    Argument parser for plethquality.

    This function creates and configures an argument parser for the plethquality
    command-line tool that calculates quality metrics from cardiac text files.

    Returns
    -------
    argparse.ArgumentParser
        Configured argument parser object with all required and optional
        arguments for plethquality functionality.

    Notes
    -----
    The parser includes both required and optional arguments for processing
    cardiac text files and generating quality metrics.

    Examples
    --------
    >>> parser = _get_parser()
    >>> args = parser.parse_args(['input.txt', 'output.txt'])
    """
    parser = argparse.ArgumentParser(
        prog="plethquality",
        description=("Calculate quality metrics from a cardiac text file."),
        allow_abbrev=False,
    )

    # Required arguments
    pf.addreqinputtextfile(parser, "infilename", onecol=True)
    parser.add_argument("outfilename", help="The name of the output text file.")

    # add optional arguments
    parser.add_argument(
        "--samplerate",
        type=lambda x: pf.is_float(parser, x),
        help="Sample rate of timecourse, in Hz (if not specified in input file).",
        default=None,
    )
    parser.add_argument(
        "--nodisplay",
        dest="display",
        action="store_false",
        help=("Do not plot the data (for noninteractive use)."),
        default=True,
    )
    return parser


def plethquality(waveform: Any, Fs: Any, S_windowsecs: float = 5.0, debug: bool = False) -> None:
    """
    Calculate the windowed skewness quality metrics for a photoplethysmogram (PPG) signal.

    This function computes the signal quality index based on the skewness of the PPG waveform
    over a sliding window, as described in Elgendi, M. "Optimal Signal Quality Index for
    Photoplethysmogram Signals". Bioengineering 2016, Vol. 3, Page 21 (2016).

    Parameters
    ----------
    waveform : array-like
        The cardiac waveform to be assessed.
    Fs : float
        The sample rate of the data in Hz.
    S_windowsecs : float, optional
        Window duration in seconds. Defaults to 5.0.
    debug : bool, optional
        Turn on extended output for debugging purposes. Defaults to False.

    Returns
    -------
    S_sqi_mean : float
        The mean value of the quality index over all time.
    S_sqi_std : float
        The standard deviation of the quality index over all time.
    S_waveform : array
        The quality metric computed over all timepoints.

    Notes
    -----
    The window size is rounded to the nearest odd number of samples to ensure symmetric
    sliding windows around each point. The skewness is calculated using `scipy.stats.skew`
    with `nan_policy="omit"` to ignore NaN values.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.stats import skew
    >>> waveform = np.random.randn(1000)
    >>> Fs = 100.0
    >>> mean_sqi, std_sqi, sqi_waveform = plethquality(waveform, Fs)
    >>> print(f"Mean SQI: {mean_sqi:.3f}")
    Mean SQI: 0.000

    """
    # calculate S_sqi over a sliding window.  Window size should be an odd number of points.
    S_windowpts = int(np.round(S_windowsecs * Fs, 0))
    S_windowpts += 1 - S_windowpts % 2
    S_waveform = np.zeros_like(waveform)
    if debug:
        print("S_windowsecs, S_windowpts:", S_windowsecs, S_windowpts)
    for i in range(0, len(waveform)):
        startpt = np.max([0, i - S_windowpts // 2])
        endpt = np.min([i + S_windowpts // 2, len(waveform)])
        S_waveform[i] = skew(waveform[startpt : endpt + 1], nan_policy="omit")
        if debug:
            print(i, startpt, endpt, endpt - startpt + 1, S_waveform[i])

    S_sqi_mean = np.mean(S_waveform)
    S_sqi_std = np.std(S_waveform)

    return S_sqi_mean, S_sqi_std, S_waveform


def plethquality(args: Any) -> None:
    """
    Calculate plethysmography quality score and optionally display results.

    This function reads plethysmography data from a text file, calculates a quality
    score based on the signal characteristics, and writes the quality scores to an
    output file. Optionally displays the quality score plot.

    Parameters
    ----------
    args : Any
        An object containing command line arguments with the following attributes:
        - infilename : str
            Input filename containing plethysmography data
        - outfilename : str
            Output filename for quality scores
        - samplerate : float, optional
            Sampling rate of the data (if not specified, will be read from file)
        - display : bool
            Whether to display the quality score plot

    Returns
    -------
    None
        This function does not return a value but writes results to files and
        optionally displays plots.

    Notes
    -----
    The function uses `tide_io.readvectorsfromtextfile` to read data and
    `tide_io.writevec` to write quality scores. Quality scores are calculated
    using an internal `plethquality` function that analyzes signal characteristics.

    Examples
    --------
    >>> args = argparse.Namespace(
    ...     infilename='pleth_data.txt',
    ...     outfilename='quality_scores.txt',
    ...     samplerate=100.0,
    ...     display=True
    ... )
    >>> plethquality(args)
    """
    if args.display:
        import matplotlib as mpl

        mpl.use("TkAgg")
        import matplotlib.pyplot as plt

    Fs, starttime, dummy, plethdata, dummy, dummy = tide_io.readvectorsfromtextfile(
        args.infilename, onecol=True
    )

    if args.samplerate is not None:
        Fs = args.samplerate
    elif Fs is None:
        print("no samplerate found in file - must be set with the --samplerate option")
        sys.exit()

    # calculate the quality score
    s_mean, s_std, quality = plethquality(plethdata, Fs)
    print(args.infilename, s_mean, "+/-", s_std)
    tide_io.writevec(quality, args.outfilename)

    if args.display:
        plt.figure()
        plt.plot(quality)
        plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
import sys
from argparse import Namespace
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray

import rapidtide.fit as tide_fit
import rapidtide.io as tide_io
from rapidtide.workflows.parser_funcs import is_valid_file


def _get_parser() -> Any:
    """
    Argument parser for polyfitim.

    This function constructs and returns an `argparse.ArgumentParser` object configured
    to parse command-line arguments for the `polyfitim` tool. It is designed to handle
    the fitting of a spatial template to 3D or 4D NIFTI files, with optional region-specific
    fitting using an atlas file.

    Returns
    -------
    argparse.ArgumentParser
        Configured argument parser for `polyfitim` command-line interface.

    Notes
    -----
    The parser expects several required NIFTI files:
    - Data file (`datafile`)
    - Data mask file (`datamask`)
    - Template file (`templatefile`)
    - Template mask file (`templatemask`)

    Optional arguments include:
    - `--regionatlas`: File containing a 3D NIFTI atlas for region-specific fitting.
    - `--order`: Polynomial order for the fit (default is 1).

    Examples
    --------
    >>> parser = _get_parser()
    >>> args = parser.parse_args(['data.nii', 'mask.nii', 'template.nii', 'template_mask.nii', 'output'])
    >>> print(args.datafile)
    'data.nii'
    """
    parser = argparse.ArgumentParser(
        prog="polyfitim",
        description="Fit a spatial template to a 3D or 4D NIFTI file.",
        allow_abbrev=False,
    )

    # Required arguments
    parser.add_argument(
        "datafile",
        type=lambda x: is_valid_file(parser, x),
        help="The name of the 3 or 4 dimensional nifti file to fit.",
    )
    parser.add_argument(
        "datamask",
        type=lambda x: is_valid_file(parser, x),
        help="The name of the 3 or 4 dimensional nifti file valid voxel mask (must match datafile).",
    )
    parser.add_argument(
        "templatefile",
        type=lambda x: is_valid_file(parser, x),
        help="The name of the 3D nifti template file (must match datafile).",
    )
    parser.add_argument(
        "templatemask",
        type=lambda x: is_valid_file(parser, x),
        help="The name of the 3D nifti template mask (must match datafile).",
    )
    parser.add_argument("outputroot", type=str, help="The root name for all output files.")
    parser.add_argument(
        "--regionatlas",
        metavar="ATLASFILE",
        dest="regionatlas",
        type=lambda x: is_valid_file(parser, x),
        help="Do individual fits to every region in ATLASFILE (3D NIFTI file).",
        default=None,
    )
    parser.add_argument(
        "--order",
        metavar="ORDER",
        dest="order",
        type=int,
        help="Perform fit to ORDERth order (default (and minimum) is 1)",
        default=1,
    )
    return parser


def polyfitim(
    datafile: Any,
    datamask: Any,
    templatefile: Any,
    templatemask: Any,
    outputroot: Any,
    regionatlas: Optional[Any] = None,
    order: int = 1,
) -> None:
    """
    Fit polynomial models to time series data within regions defined by masks.

    This function performs polynomial fitting of specified order to time series data
    within spatial regions defined by a template mask. It supports both global and
    region-specific fitting, and outputs fitted time series, polynomial coefficients,
    and R² values.

    Parameters
    ----------
    datafile : Any
        Path to the input NIfTI file containing the time series data.
    datamask : Any
        Path to the NIfTI file containing the data mask. Can be 3D or 4D.
    templatefile : Any
        Path to the NIfTI file containing the template (e.g., a regressor) for fitting.
    templatemask : Any
        Path to the NIfTI file containing the mask for the template.
    outputroot : Any
        Root name for output files (e.g., 'output' will produce 'output_fit.nii.gz').
    regionatlas : Any, optional
        Path to the NIfTI file containing region labels. If provided, fitting is
        performed separately for each region. Default is None.
    order : int, optional
        Order of the polynomial to fit. Must be >= 1. Default is 1 (linear fit).

    Returns
    -------
    None
        Function writes multiple output files:
        - Fitted time series (NIfTI format)
        - Residuals (NIfTI format)
        - R² values (text file)
        - Polynomial coefficients (text files)

    Notes
    -----
    - The function assumes that all input files are in NIfTI format.
    - If `datamask` is 4D, it is treated as a time-varying mask.
    - If `regionatlas` is provided, fitting is performed separately for each region.
    - The function uses `tide_io` for reading/writing NIfTI files and `tide_fit.mlregress`
      for polynomial regression.

    Examples
    --------
    >>> polyfitim(
    ...     datafile='data.nii.gz',
    ...     datamask='mask.nii.gz',
    ...     templatefile='template.nii.gz',
    ...     templatemask='template_mask.nii.gz',
    ...     outputroot='output',
    ...     regionatlas='atlas.nii.gz',
    ...     order=2
    ... )
    """
    # check the order
    if order < 1:
        print("order must be >= 1")
        sys.exit()

    # read in data
    print("reading in data arrays")
    (
        datafile_img,
        datafile_data,
        datafile_hdr,
        datafiledims,
        datafilesizes,
    ) = tide_io.readfromnifti(datafile)
    (
        datamask_img,
        datamask_data,
        datamask_hdr,
        datamaskdims,
        datamasksizes,
    ) = tide_io.readfromnifti(datamask)
    (
        templatefile_img,
        templatefile_data,
        templatefile_hdr,
        templatefiledims,
        templatefilesizes,
    ) = tide_io.readfromnifti(templatefile)
    (
        templatemask_img,
        templatemask_data,
        templatemask_hdr,
        templatemaskdims,
        templatemasksizes,
    ) = tide_io.readfromnifti(templatemask)

    if regionatlas is not None:
        (
            regionatlas_img,
            regionatlas_data,
            regionatlas_hdr,
            regionatlasdims,
            regionatlassizes,
        ) = tide_io.readfromnifti(regionatlas)

    xsize = datafiledims[1]
    ysize = datafiledims[2]
    numslices = datafiledims[3]
    timepoints = datafiledims[4]

    # check dimensions
    print("checking dimensions")
    if not tide_io.checkspacedimmatch(datafiledims, datamaskdims):
        print("input mask spatial dimensions do not match image")
        exit()
    if datamaskdims[4] == 1:
        print("using 3d data mask")
        datamask3d = True
    else:
        datamask3d = False
        if not tide_io.checktimematch(datafiledims, datamaskdims):
            print("input mask time dimension does not match image")
            exit()
    if not tide_io.checkspacedimmatch(datafiledims, templatefiledims):
        print(templatefiledims, "template file spatial dimensions do not match image")
        exit()
    if not templatefiledims[4] == 1:
        print("template file time dimension is not equal to 1")
        exit()
    if not tide_io.checkspacedimmatch(datafiledims, templatemaskdims):
        print("template mask spatial dimensions do not match image")
        exit()
    if not templatemaskdims[4] == 1:
        print("template mask time dimension is not equal to 1")
        exit()
    if regionatlas is not None:
        if not tide_io.checkspacedimmatch(datafiledims, regionatlasdims):
            print("template mask spatial dimensions do not match image")
            exit()
        if not regionatlasdims[4] == 1:
            print("regionatlas time dimension is not equal to 1")
            exit()

    # allocating arrays
    print("allocating arrays")
    numspatiallocs = int(xsize) * int(ysize) * int(numslices)
    rs_datafile = datafile_data.reshape((numspatiallocs, timepoints))
    if datamask3d:
        rs_datamask = datamask_data.reshape(numspatiallocs)
    else:
        rs_datamask = datamask_data.reshape((numspatiallocs, timepoints))
    rs_datamask_bin = np.where(rs_datamask > 0.9, 1.0, 0.0)
    rs_templatefile = templatefile_data.reshape(numspatiallocs)
    rs_templatemask = templatemask_data.reshape(numspatiallocs)
    rs_templatemask_bin = np.where(rs_templatemask > 0.1, 1.0, 0.0)
    if regionatlas is not None:
        rs_regionatlas = regionatlas_data.reshape(numspatiallocs)
        numregions = int(np.max(rs_regionatlas))
        print(f"atlas file {regionatlas} has {numregions} regions.")

    fitdata = np.zeros((numspatiallocs, timepoints), dtype="float")
    if regionatlas is not None:
        polycoffs = np.zeros((numregions, order + 1, timepoints), dtype="float")
        r2vals = np.zeros((numregions, timepoints), dtype="float")
    else:
        polycoffs = np.zeros((order + 1, timepoints), dtype="float")
        r2vals = np.zeros(timepoints, dtype="float")

    if regionatlas is not None:
        print("making region masks")
        regionvoxels = np.zeros((numspatiallocs, numregions), dtype="float")
        for region in range(0, numregions):
            thisregion = np.where((rs_regionatlas * rs_templatemask_bin) == (region + 1))
            regionvoxels[thisregion, region] = 1.0

    # cycle over all images
    print("now cycling over all images")
    for thetime in range(0, timepoints):
        print("fitting timepoint", thetime)

        # get the appropriate mask
        if datamask3d:
            for i in range(timepoints):
                thisdatamask = rs_datamask_bin
        else:
            thisdatamask = rs_datamask_bin[:, thetime]
        if regionatlas is not None:
            for region in range(0, numregions):
                voxelstofit = np.where(regionvoxels[:, region] * thisdatamask > 0.5)
                voxelstoreconstruct = np.where(regionvoxels[:, region] > 0.5)
                evlist = []
                for i in range(1, order + 1):
                    evlist.append((rs_templatefile[voxelstofit]) ** i)
                thefit, R2 = tide_fit.mlregress(
                    evlist,
                    rs_datafile[voxelstofit, thetime][0],
                )
                for i in range(order + 1):
                    polycoffs[region, i, thetime] = thefit[0, i]
                fitdata[voxelstoreconstruct, thetime] = polycoffs[region, 0, thetime]
                for i in range(1, order + 1):
                    fitdata[voxelstoreconstruct, thetime] += polycoffs[region, i, thetime] * (
                        rs_templatefile[voxelstoreconstruct] ** i
                    )
                r2vals[region, thetime] = R2
        else:
            voxelstofit = np.where(thisdatamask > 0.5)
            voxelstoreconstruct = np.where(rs_templatemask > 0.5)
            evlist = []
            for i in range(1, order + 1):
                evlist.append((rs_templatefile[voxelstofit]) ** i)
            thefit, R2 = tide_fit.mlregress(evlist, rs_datafile[voxelstofit, thetime][0])
            for i in range(order + 1):
                polycoffs[i, thetime] = thefit[0, i]
            fitdata[voxelstoreconstruct, thetime] = polycoffs[0, thetime]
            for i in range(1, order + 1):
                fitdata[voxelstoreconstruct, thetime] += polycoffs[i, thetime] * (
                    rs_templatefile[voxelstoreconstruct] ** i
                )
            r2vals[thetime] = R2
    residuals = rs_datafile - fitdata

    # write out the data files
    print("writing time series")

    tide_io.writenpvecs(r2vals, outputroot + "_r2vals.txt")
    if regionatlas is not None:
        for region in range(0, numregions):
            outstring = f"region {region + 1}:"
            for i in range(order + 1):
                tide_io.writenpvecs(polycoffs[i, :], outputroot + "_O_i_coffs.txt")
                outstring += f" O_{i} mean, std {np.mean(polycoffs[region, i, :])}, {np.std(polycoffs[region, i, :])}"
            print(outstring)
    else:
        outstring = ""
        for i in range(order + 1):
            tide_io.writenpvecs(polycoffs[:], outputroot + f"_O_{i}_coffs.txt")
            outstring += f" O_{i} mean, std {np.mean(polycoffs[i, :])}, {np.std(polycoffs[i, :])}"
        print(outstring)

    print("writing nifti series")
    tide_io.savetonifti(
        fitdata.reshape((xsize, ysize, numslices, timepoints)),
        datafile_hdr,
        outputroot + "_fit",
    )
    tide_io.savetonifti(
        residuals.reshape((xsize, ysize, numslices, timepoints)),
        datafile_hdr,
        outputroot + "_residuals",
    )


def main(args: Any) -> None:
    """
    Main function to perform polynomial fitting on imaging data.

    This function serves as the entry point for polynomial fitting operations
    on medical imaging data using template-based registration and fitting.

    Parameters
    ----------
    args : Any
        Namespace object containing all required arguments for the polynomial fitting
        operation. Expected attributes include:

        - datafile : str
            Path to the input data file to be fitted
        - datamask : str
            Path to the data mask file for region of interest specification
        - templatefile : str
            Path to the template file for reference
        - templatemask : str
            Path to the template mask file for reference region specification
        - outputroot : str
            Root path for output files
        - regionatlas : str, optional
            Path to region atlas file for anatomical labeling
        - order : int, optional
            Order of the polynomial to fit (default is typically 1)

    Returns
    -------
    None
        This function does not return any value. It performs the polynomial fitting
        operation and saves results to the specified output directory.

    Notes
    -----
    The function internally calls `polyfitim` with the provided arguments to
    perform the actual polynomial fitting computation. All input files must be
    properly formatted and accessible.

    Examples
    --------
    >>> import argparse
    >>> args = argparse.Namespace(
    ...     datafile='data.nii.gz',
    ...     datamask='data_mask.nii.gz',
    ...     templatefile='template.nii.gz',
    ...     templatemask='template_mask.nii.gz',
    ...     outputroot='output',
    ...     regionatlas='atlas.nii.gz',
    ...     order=2
    ... )
    >>> main(args)
    """
    polyfitim(
        args.datafile,
        args.datamask,
        args.templatefile,
        args.templatemask,
        args.outputroot,
        regionatlas=args.regionatlas,
        order=args.order,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
import sys
from argparse import Namespace
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import nibabel as nib
import numpy as np
import pandas as pd
from numpy.polynomial import Polynomial
from numpy.typing import NDArray

import rapidtide.correlate as tide_corr
import rapidtide.fit as tide_fit
import rapidtide.io as tide_io
import rapidtide.miscmath as tide_math
import rapidtide.resample as tide_resamp
import rapidtide.workflows.parser_funcs as pf


def _get_parser() -> Any:
    """
    Create and configure an argument parser for the showstxcorr tool.

    This function sets up an `argparse.ArgumentParser` with various required and
    optional arguments to control the behavior of the cross-correlation plotting
    tool. It includes options for input files, sampling parameters, correlation
    settings, display control, and preprocessing steps.

    Returns
    -------
    argparse.ArgumentParser
        Configured argument parser object with all necessary arguments added.

    Notes
    -----
    The parser is configured with:
    - Two required input text files (`infilename1`, `infilename2`)
    - Output filename root
    - Sampling rate or time step options (mutually exclusive)
    - Correlation threshold and window parameters
    - Time range and display options
    - Preprocessing options like detrending and correlation weighting
    - Filtering and windowing options

    Examples
    --------
    >>> parser = _get_parser()
    >>> args = parser.parse_args()
    """
    parser = argparse.ArgumentParser(
        prog="showstxcorr",
        description="Plots the data in text files.",
        allow_abbrev=False,
    )

    pf.addreqinputtextfile(parser, "infilename1", onecol=True)
    pf.addreqinputtextfile(parser, "infilename2", onecol=True)
    parser.add_argument(
        "outfilename",
        type=str,
        action="store",
        help="The root name of the output files.",
    )

    sampling = parser.add_mutually_exclusive_group()
    sampling.add_argument(
        "--samplerate",
        dest="samplerate",
        action="store",
        metavar="FREQ",
        type=lambda x: pf.is_float(parser, x),
        help=(
            "Set the sample rate of the data file to FREQ. "
            "If neither samplerate or sampletime is specified, sample rate is 1.0."
        ),
        default="auto",
    )
    sampling.add_argument(
        "--sampletime",
        dest="samplerate",
        action="store",
        metavar="TSTEP",
        type=lambda x: pf.invert_float(parser, x),
        help=(
            "Set the sample rate of the data file to 1.0/TSTEP. "
            "If neither samplerate or sampletime is specified, sample rate is 1.0."
        ),
        default="auto",
    )
    # add optional arguments
    parser.add_argument(
        "--corrthresh",
        dest="corrthresh",
        metavar="THRESH",
        type=lambda x: pf.is_float(parser, x),
        help=("Cross correlation magnitude threshold to accept a delay value (default is 0.5)."),
        default=0.5,
    )
    parser.add_argument(
        "--windowwidth",
        dest="windowwidth",
        metavar="WINDOWWIDTH",
        type=lambda x: pf.is_float(parser, x),
        help="Use a window width of WINDOWWIDTH seconds (default is 50.0s).",
        default=50.0,
    )
    parser.add_argument(
        "--stepsize",
        dest="stepsize",
        metavar="STEPSIZE",
        type=lambda x: pf.is_float(parser, x),
        help=(
            "Timestep between subsequent measurements (default is 25.0s).  "
            "Will be rounded to the nearest sample time."
        ),
        default=25.0,
    )
    parser.add_argument(
        "--starttime",
        dest="starttime",
        metavar="START",
        type=float,
        help="Start plotting at START seconds (default is 0.0).",
        default=0.0,
    )
    parser.add_argument(
        "--duration",
        dest="duration",
        metavar="DURATION",
        type=float,
        help="Amount of data, in seconds, to process after starttime (default is the entire timecourse).",
        default=1000000.0,
    )
    parser.add_argument(
        "--matrixoutput",
        dest="matrixoutput",
        action="store_true",
        help=("Format output files as NIFTI files and .csv matrices (default is text files)."),
        default=False,
    )
    parser.add_argument(
        "--nodisplay",
        dest="display",
        action="store_false",
        help=("Do not plot the data (for noninteractive use)"),
        default=True,
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help=("Enable additional debugging output."),
        default=False,
    )
    parser.add_argument(
        "--verbose",
        dest="verbose",
        action="store_true",
        help=("Print out more debugging information"),
        default=False,
    )
    pf.addsearchrangeopts(parser, details=True, defaultmin=-15.0, defaultmax=15.0)
    pf.addtimerangeopts(parser)

    preproc = parser.add_argument_group()
    preproc.add_argument(
        "--detrendorder",
        dest="detrendorder",
        action="store",
        type=int,
        metavar="ORDER",
        help=("Set order of trend removal (0 to disable, default is 1 - linear). "),
        default=1,
    )
    # add window options
    pf.addwindowopts(parser)

    # Filter arguments
    pf.addfilteropts(parser, filtertarget="timecourses", details=True)

    # Preprocessing options
    preproc = parser.add_argument_group("Preprocessing options")
    preproc.add_argument(
        "--corrweighting",
        dest="corrweighting",
        action="store",
        type=str,
        choices=["None", "phat", "liang", "eckart"],
        help=("Method to use for cross-correlation " "weighting. Default is  None. "),
        default="None",
    )
    preproc.add_argument(
        "--invert",
        dest="invert",
        action="store_true",
        help=("Invert second timecourse prior to correlation. "),
        default=False,
    )
    preproc.add_argument(
        "--label",
        dest="label",
        metavar="LABEL",
        action="store",
        type=str,
        help=("Label for the delay value. "),
        default="None",
    )
    return parser


def printthresholds(pcts: Any, thepercentiles: Any, labeltext: Any) -> None:
    """
    Print thresholds with corresponding percentile values.

    This function prints a formatted list of thresholds where each threshold
    corresponds to a specific percentile. The output shows the probability
    threshold for which the percentile is less than the given value.

    Parameters
    ----------
    pcts : Any
        Array-like object containing the percentile values to be printed.
    thepercentiles : Any
        Array-like object containing the corresponding percentile thresholds.
    labeltext : Any
        Text label to be printed before the threshold values.

    Returns
    -------
    None
        This function prints to stdout and does not return any value.

    Notes
    -----
    The function assumes that both `pcts` and `thepercentiles` have the same length
    and that the percentiles are ordered in ascending order.

    Examples
    --------
    >>> pcts = [0.05, 0.01, 0.001]
    >>> thepercentiles = [0.95, 0.99, 0.999]
    >>> labeltext = "Significance thresholds:"
    >>> printthresholds(pcts, thepercentiles, labeltext)
    Significance thresholds:
            p < 0.05 :  0.05
            p < 0.01 :  0.01
            p < 0.001 :  0.001
    """
    print(labeltext)
    for i in range(0, len(pcts)):
        print("\tp <", 1.0 - thepercentiles[i], ": ", pcts[i])


def showstxcorr(args: Any) -> None:
    """
    Compute and display short-term cross-correlations between two time series.

    This function performs short-term cross-correlation analysis on two input time series,
    applying optional filtering, detrending, and time-warping. It supports both single
    correlation output and matrix output modes, and can generate plots and save results
    to files.

    Parameters
    ----------
    args : argparse.Namespace
        Command-line arguments containing the following attributes:
        - infilename1 : str
            Path to the first input file.
        - infilename2 : str
            Path to the second input file.
        - samplerate : float or str
            Sampling rate of the data. Must be set if 'auto'.
        - starttime : float
            Start time for data trimming.
        - duration : float
            Duration of the data to process.
        - stepsize : float
            Step size for correlation computation.
        - windowwidth : float
            Width of the analysis window in seconds.
        - lagmin : float
            Minimum lag for correlation search.
        - lagmax : float
            Maximum lag for correlation search.
        - corrweighting : str
            Type of weighting for correlation.
        - windowfunc : str
            Window function to apply.
        - detrendorder : int
            Order of detrending to apply.
        - invert : bool
            Whether to invert the second signal.
        - display : bool
            Whether to display plots.
        - debug : bool
            Whether to enable debug output.
        - outfilename : str
            Base name for output files.
        - corrthresh : float
            Threshold for correlation significance.
        - filter : dict
            Filter parameters.

    Returns
    -------
    None
        This function does not return a value but saves output files and displays plots
        if requested.

    Notes
    -----
    The function applies a prefilter to the data and trims the input time series to
    the specified duration. It supports both single correlation and matrix correlation
    modes. In matrix mode, all components are correlated against each other and
    results are saved to NIfTI files and CSV tables. In single mode, correlations are
    computed between two time series and results are saved to text files.

    Examples
    --------
    >>> import argparse
    >>> args = argparse.Namespace(
    ...     infilename1='data1.txt',
    ...     infilename2='data2.txt',
    ...     samplerate=100,
    ...     starttime=0,
    ...     duration=1000,
    ...     stepsize=10,
    ...     windowwidth=20,
    ...     lagmin=-5,
    ...     lagmax=5,
    ...     corrweighting='uniform',
    ...     windowfunc='hanning',
    ...     detrendorder=1,
    ...     invert=False,
    ...     display=False,
    ...     debug=False,
    ...     outfilename='output',
    ...     corrthresh=0.3,
    ...     filter={}
    ... )
    >>> showstxcorr(args)
    """
    # get the command line parameters and
    # finish up processing arguments, do initial filter setup
    args, theprefilter = pf.postprocessfilteropts(args)
    args = pf.postprocesssearchrangeopts(args)
    args = pf.postprocesstimerangeopts(args)

    if args.display:
        import matplotlib as mpl

        mpl.use("TkAgg")
        import matplotlib.pyplot as plt

    # check that required arguments are set
    if args.samplerate == "auto":
        print("samplerate must be set")
        sys.exit()
    sampletime = 1.0 / args.samplerate

    # read in the files and get everything trimmed to the right length
    stepsize = sampletime * np.round(args.stepsize / sampletime)

    # Now update the lower limit of the filter.
    init_lowerstop, init_lowerpass, init_upperpass, init_upperstop = theprefilter.getfreqs()
    window_lowestfreq = 1.0 / args.windowwidth
    if window_lowestfreq > init_lowerpass:
        print(
            f"resetting lower limit of filter from {init_lowerpass} to {window_lowestfreq}Hz for window length {args.windowwidth}s"
        )
        theprefilter.settype("arb")
        theprefilter.setfreqs(
            window_lowestfreq,
            window_lowestfreq,
            init_upperpass,
            init_upperstop,
        )

    startpoint = max([int(args.starttime * args.samplerate), 0])
    inputdata1 = tide_io.readvec(args.infilename1)
    numpoints = len(inputdata1)
    inputdata2 = tide_io.readvec(args.infilename2)
    endpoint1 = min(
        [
            startpoint + int(args.duration * args.samplerate),
            int(len(inputdata1)),
            int(len(inputdata2)),
        ]
    )
    endpoint2 = min(
        [int(args.duration * args.samplerate), int(len(inputdata1)), int(len(inputdata2))]
    )
    trimlen = min(endpoint1 - startpoint, endpoint2)
    trimmeddata = np.zeros((2, trimlen), dtype="float")
    trimmeddata[0, :] = inputdata1[startpoint:startpoint + trimlen]
    trimmeddata[1, :] = inputdata2[0:trimlen]

    # band limit the regressors if that is needed
    if theprefilter.gettype() != "None":
        if args.verbose:
            print("filtering to ", theprefilter.gettype(), " band")

    thedims = trimmeddata.shape
    tclen = thedims[1]
    numcomponents = thedims[0]
    reformdata = np.reshape(trimmeddata, (numcomponents, tclen))

    print("preprocessing all timecourses")
    for component in range(0, numcomponents):
        filtereddata = tide_math.corrnormalize(
            theprefilter.apply(args.samplerate, reformdata[component, :]),
            windowfunc="None",
            detrendorder=args.detrendorder,
        )
        reformdata[component, :] = tide_math.stdnormalize(
            tide_fit.detrend(tide_math.stdnormalize(filtereddata), order=args.detrendorder)
        )

    xcorr_x = np.r_[0.0:tclen] * sampletime - (tclen * sampletime) / 2.0
    searchstart = int(int(tclen) // 2 + (args.lagmin * args.samplerate))
    searchend = int(int(tclen) // 2 + (args.lagmax * args.samplerate))
    xcorr_x_trim = xcorr_x[searchstart:searchend]
    if args.invert:
        flipfac = -1.0
    else:
        flipfac = 1.0

    # now that we have all the information, we have a couple of places to go:
    # We are either doing short term correlations or full timecourse
    # We are either doing two time courses from different (or the same) files, or we are doing more than 2

    if args.matrixoutput:
        # find the lengths of the outputfiles
        print("finding timecourse lengths")
        times, corrpertime, ppertime = tide_corr.shorttermcorr_1D(
            reformdata[0, :],
            reformdata[0, :],
            sampletime,
            args.windowwidth,
            samplestep=int(stepsize // sampletime),
            windowfunc=args.windowfunc,
            detrendorder=0,
        )
        plength = len(times)
        times, xcorrpertime, Rvals, delayvals, valid = tide_corr.shorttermcorr_2D(
            reformdata[0, :],
            reformdata[0, :],
            sampletime,
            args.windowwidth,
            laglimits=[args.lagmin, args.lagmax],
            samplestep=int(stepsize // sampletime),
            weighting=args.corrweighting,
            windowfunc=args.windowfunc,
            detrendorder=0,
            displayplots=False,
        )
        xlength = len(times)

        # now allocate the output arrays
        print("allocating data arrays")
        Rvals = np.zeros((numcomponents, numcomponents, 1, xlength), dtype="float")
        delayvals = np.zeros((numcomponents, numcomponents, 1, xlength), dtype="float")
        valid = np.zeros((numcomponents, numcomponents, 1, xlength), dtype="float")
        corrpertime = np.zeros((numcomponents, numcomponents, 1, plength), dtype="float")
        ppertime = np.zeros((numcomponents, numcomponents, 1, plength), dtype="float")

        # do the correlations
        for component1 in range(0, numcomponents):
            print("correlating with component", component1)
            for component2 in range(0, numcomponents):
                (
                    times,
                    corrpertime[component1, component2, 0, :],
                    ppertime[component1, component2, 0, :],
                ) = tide_corr.shorttermcorr_1D(
                    reformdata[component1, :],
                    flipfac * reformdata[component2, :],
                    sampletime,
                    args.windowwidth,
                    samplestep=int(stepsize // sampletime),
                    windowfunc=args.windowfunc,
                    detrendorder=0,
                )
                (
                    times,
                    xcorrpertime,
                    Rvals[component1, component2, 0, :],
                    delayvals[component1, component2, 0, :],
                    valid[component1, component2, 0, :],
                ) = tide_corr.shorttermcorr_2D(
                    reformdata[component1, :],
                    flipfac * reformdata[component2, :],
                    sampletime,
                    args.windowwidth,
                    laglimits=[args.lagmin, args.lagmax],
                    samplestep=int(stepsize // sampletime),
                    weighting=args.corrweighting,
                    windowfunc=args.windowfunc,
                    detrendorder=0,
                    displayplots=False,
                )

        outputaffine = np.eye(4)
        # input_img, input_data, input_hdr, thedims, thesizes = tide_io.readfromnifti(inputfilename)
        init_img = nib.Nifti1Image(corrpertime, outputaffine)
        init_hdr = init_img.header.copy()
        init_sizes = init_hdr["pixdim"].copy()
        init_sizes[4] = sampletime
        init_hdr["toffset"] = times[0]
        tide_io.savetonifti(corrpertime, init_hdr, args.outfilename + "_pearsonR")
        tide_io.savetonifti(ppertime, init_hdr, args.outfilename + "_corrp")
        tide_io.savetonifti(Rvals, init_hdr, args.outfilename + "_maxxcorr")
        tide_io.savetonifti(delayvals, init_hdr, args.outfilename + "_delayvals")
        tide_io.savetonifti(valid, init_hdr, args.outfilename + "_valid")
        rows = []
        cols = []
        for i in range(numcomponents):
            rows.append("region " + str(i + 1))
            cols.append("region " + str(i + 1))
        for segment in range(plength):
            df = pd.DataFrame(data=corrpertime[:, :, 0, 0], columns=cols)
            df.insert(0, "sources", pd.Series(rows))
            df.to_csv(
                args.outfilename + "_seg_" + str(segment).zfill(4) + "_pearsonR.csv",
                index=False,
            )
            df = pd.DataFrame(data=ppertime[:, :, 0, 0], columns=cols)
            df.insert(0, "sources", pd.Series(rows))
            df.to_csv(
                args.outfilename + "_seg_" + str(segment).zfill(4) + "_corrp.csv",
                index=False,
            )
        for segment in range(xlength):
            df = pd.DataFrame(data=Rvals[:, :, 0, 0], columns=cols)
            df.insert(0, "sources", pd.Series(rows))
            df.to_csv(
                args.outfilename + "_seg_" + str(segment).zfill(4) + "_maxxcorr.csv",
                index=False,
            )
            df = pd.DataFrame(data=delayvals[:, :, 0, 0], columns=cols)
            df.insert(0, "sources", pd.Series(rows))
            df.to_csv(
                args.outfilename + "_seg_" + str(segment).zfill(4) + "_delayvals.csv",
                index=False,
            )
            df = pd.DataFrame(data=valid[:, :, 0, 0], columns=cols)
            df.insert(0, "sources", pd.Series(rows))
            df.to_csv(
                args.outfilename + "_seg_" + str(segment).zfill(4) + "_valid.csv",
                index=False,
            )

    else:
        times, corrpertime, ppertime = tide_corr.shorttermcorr_1D(
            reformdata[0, :],
            flipfac * reformdata[1, :],
            sampletime,
            args.windowwidth,
            samplestep=int(stepsize // sampletime),
            windowfunc=args.windowfunc,
            detrendorder=0,
        )
        times, xcorrpertime, Rvals, delayvals, valid = tide_corr.shorttermcorr_2D(
            reformdata[0, :],
            flipfac * reformdata[1, :],
            sampletime,
            args.windowwidth,
            laglimits=[args.lagmin, args.lagmax],
            samplestep=int(stepsize // sampletime),
            weighting=args.corrweighting,
            windowfunc=args.windowfunc,
            detrendorder=0,
            displayplots=False,
        )
        tide_io.writenpvecs(corrpertime, args.outfilename + "_pearson.txt")
        tide_io.writenpvecs(ppertime, args.outfilename + "_pvalue.txt")
        tide_io.writenpvecs(Rvals, args.outfilename + "_Rvalue.txt")
        tide_io.writenpvecs(delayvals, args.outfilename + "_delay.txt")
        tide_io.writenpvecs(valid, args.outfilename + "_mask.txt")

        # threshold by corrthresh
        validlocs = np.where(np.fabs(Rvals) > args.corrthresh)

        # do a polynomial fit to the delay function
        thefit = (
            Polynomial.fit(
                times[validlocs],
                delayvals[validlocs],
                7,
                w=(Rvals[validlocs] * Rvals[validlocs]),
            )
            .convert()
            .coef[::-1]
        )
        smoothdelayvals = np.poly1d(thefit)(times)

        hirestimeaxis = np.linspace(0, sampletime * endpoint2, num=endpoint2, endpoint=False)
        highresdelayvals = np.poly1d(thefit)(hirestimeaxis)
        if args.debug:
            print("len(hirestimeaxis):", len(hirestimeaxis))
            print("len(highresdelayvals):", len(highresdelayvals))
            print("len(reformdata[1, :]):", len(reformdata[1, :]))

        timewarped = tide_resamp.timewarp(
            hirestimeaxis,
            reformdata[1, :],
            highresdelayvals,
            debug=args.debug,
        )
        tide_io.writenpvecs(timewarped, args.outfilename + "_timewarped.txt")
        tide_io.writenpvecs(highresdelayvals, args.outfilename + "_hiresdelayvals.txt")

        if args.display:
            # timeaxis = np.r_[0.0:len(filtereddata1)] * sampletime
            fig, ax1 = plt.subplots()
            ax1.plot(times, corrpertime, "k")
            ax1.set_ylabel("Pearson R", color="k")
            ax2 = ax1.twinx()
            ax2.plot(times, ppertime, "r")
            ax2.set_ylabel("p value", color="r")
            fig, ax3 = plt.subplots()
            ax3.plot(times[validlocs], Rvals[validlocs], "k")
            ax3.set_ylabel("Xcorr max R", color="k")
            ax4 = ax3.twinx()
            ax4.plot(times[validlocs], delayvals[validlocs], marker=".", linestyle="None")
            ax4.plot(times, smoothdelayvals, "g")
            ax4.set_ylabel("Delay (s)", color="r")
            plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from rapidtide.filter import dolpfiltfilt
from rapidtide.resample import timeshift
from rapidtide.tests.utils import mse


def test_timeshift(debug=False):
    tr = 1.0
    padtime = 300.0
    testlen = 1000
    shiftdist = 30
    timeaxis = np.arange(0.0, 1.0 * testlen) * tr
    # timecoursein = np.zeros((testlen), dtype='float64')
    timecoursein = np.zeros_like(timeaxis, np.float64)
    midpoint = int(testlen // 2) + 1
    timecoursein[midpoint - 1] = np.float64(1.0)
    timecoursein[midpoint] = np.float64(1.0)
    timecoursein[midpoint + 1] = np.float64(1.0)
    timecoursein -= 0.5
    butterorder = 4
    timecoursein = 0.5 * dolpfiltfilt(1.0, 0.25, timecoursein, butterorder) + 0.5

    shiftlist = [-30, -20, -10, 0, 10, 20, 30]
    shiftlist = [-100]

    if debug:
        plt.figure()
        plt.ylim([-1.0, 2.0 * len(shiftlist) + 1.0])
        plt.plot(timecoursein)
        legend = ["Original"]
        offset = 0.0

    for shiftdist in shiftlist:
        # generate the ground truth rolled regressor
        tcrolled = np.float64(np.roll(timecoursein, int(shiftdist)))

        # generate the fast resampled regressor
        # tcshifted = genlaggedtc.yfromx(timeaxis - shiftdist, debug=debug)
        tcshifted, weights, alltc, allweights = timeshift(
            timecoursein, 1.0 * shiftdist, int(padtime // tr), doplot=False
        )

        # print out all elements
        print("len tcrolled=", len(tcrolled), "len tcshifted=", len(tcshifted))
        for i in range(0, len(tcrolled)):
            print(i, tcrolled[i], tcshifted[i], tcshifted[i] - tcrolled[i])

        # plot if we are doing that
        if debug:
            offset += 1.0
            plt.plot(tcrolled + offset)
            legend.append("Roll " + str(shiftdist))
            offset += 1.0
            plt.plot(tcshifted + offset)
            legend.append("Timeshift " + str(shiftdist))
            plt.plot(weights + offset)
            legend.append("Weights " + str(shiftdist))

        # do the tests
        msethresh = 1e-6
        aethresh = 2
        assert mse(tcrolled, tcshifted) < msethresh
        np.testing.assert_almost_equal(tcrolled, tcshifted, aethresh)

    if debug:
        plt.legend(legend)
        plt.show()


if __name__ == "__main__":
    mpl.use("TkAgg")
    test_timeshift(debug=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
from typing import Any, Tuple

import numpy as np
from numpy.typing import NDArray

import rapidtide.refinedelay as tide_refinedelay
import rapidtide.stats as tide_stats


def refineDelay(
    fmri_data_valid: NDArray,
    initial_fmri_x: NDArray,
    xdim: float,
    ydim: float,
    slicethickness: float,
    sLFOfiltmask: NDArray,
    genlagtc: Any,
    oversamptr: float,
    sLFOfitmean: NDArray,
    rvalue: NDArray,
    r2value: NDArray,
    fitNorm: Any,
    fitcoeff: NDArray,
    lagtc: NDArray,
    outputname: str,
    validvoxels: Any,
    nativespaceshape: Any,
    theinputdata: Any,
    lagtimes: Any,
    optiondict: Any,
    LGR: Any,
    TimingLGR: Any,
    outputlevel: str = "normal",
    gausssigma: int = -1,
    patchthresh: float = 3.0,
    timemask: NDArray | None = None,
    mindelay: float = -5.0,
    maxdelay: float = 5.0,
    numpoints: int = 501,
    histlen: int = 101,
    rt_floattype: np.dtype = np.dtype(np.float64),
    debug: bool = False,
) -> Tuple[NDArray, NDArray, NDArray, NDArray, float]:
    """
    Refine delay estimates using regression derivative ratios and histogram-based calibration.

    This function performs calibration of delay estimates by computing regression derivative
    ratios, filtering them, training a mapping from ratios to delays, and finally calculating
    delay offsets for each voxel. It also generates a histogram of the computed delay offsets.

    Parameters
    ----------
    fmri_data_valid : Any
        Valid fMRI data used for delay refinement.
    initial_fmri_x : Any
        Initial fMRI design matrix.
    xdim : Any
        X dimension of the data.
    ydim : Any
        Y dimension of the data.
    slicethickness : Any
        Thickness of the slices in the data.
    sLFOfiltmask : NDArray
        Mask for filtering based on SLOF (slice timing) effects.
    genlagtc : Any
        Generated lag time course.
    oversamptr : Any
        Oversampling time resolution.
    sLFOfitmean : Any
        Mean SLOF fit values.
    rvalue : Any
        R-values from regression.
    r2value : Any
        R-squared values from regression.
    fitNorm : Any
        Normalized fit coefficients.
    fitcoeff : Any
        Fit coefficients.
    lagtc : Any
        Lag time course.
    outputname : Any
        Base name for output files.
    validvoxels : Any
        Indices of valid voxels.
    nativespaceshape : Any
        Shape of the native space.
    theinputdata : Any
        Input data object.
    lagtimes : Any
        Time lags used for analysis.
    optiondict : Any
        Dictionary of options for processing.
    LGR : Any
        Logger for general messages.
    TimingLGR : Any
        Logger for timing-related messages.
    outputlevel : str, optional
        Level of output verbosity, default is "normal".
    gausssigma : int, optional
        Sigma for Gaussian filtering, default is -1 (no filtering).
    patchthresh : float, optional
        Threshold for patch-based filtering, default is 3.0.
    timemask : NDArray, optional
       Mask of timepoints to include in regression filtering. Default is None.
    mindelay : float, optional
        Minimum delay value, default is -5.0.
    maxdelay : float, optional
        Maximum delay value, default is 5.0.
    numpoints : int, optional
        Number of points for delay interpolation, default is 501.
    histlen : int, optional
        Length of histogram bins, default is 101.
    rt_floattype : np.dtype, optional
        Data type for rapidtide float operations, default is np.float64.
    debug : bool, optional
        Enable debug mode, default is False.

    Returns
    -------
    tuple
        A tuple containing:
        - delayoffset : ndarray
            Calculated delay offsets for each voxel.
        - regressderivratios : ndarray
            Raw regression derivative ratios.
        - medfiltregressderivratios : ndarray
            Median-filtered regression derivative ratios.
        - filteredregressderivratios : ndarray
            Final filtered regression derivative ratios.
        - delayoffsetMAD : ndarray
            Median absolute deviation of delay offsets.

    Notes
    -----
    The function uses the `tide_refinedelay` module to perform various steps:
    1. Computes regression derivative ratios using `getderivratios`.
    2. Filters these ratios using `filterderivratios`.
    3. Trains a ratio-to-delay mapping using `trainratiotooffset`.
    4. Converts ratios to delay offsets using `ratiotodelay`.
    5. Saves a histogram of delay offsets to disk.

    Examples
    --------
    >>> delayoffset, regressderivratios, medfiltregressderivratios, filteredregressderivratios, delayoffsetMAD = refineDelay(
    ...     fmri_data_valid, initial_fmri_x, xdim, ydim, slicethickness,
    ...     sLFOfiltmask, genlagtc, oversamptr, sLFOfitmean, rvalue, r2value,
    ...     fitNorm, fitcoeff, lagtc, outputname, validvoxels, nativespaceshape,
    ...     theinputdata, lagtimes, optiondict, LGR, TimingLGR
    ... )
    """
    # do the calibration
    TimingLGR.info("Refinement calibration start")
    regressderivratios, regressrvalues = tide_refinedelay.getderivratios(
        fmri_data_valid,
        validvoxels,
        initial_fmri_x,
        lagtimes,
        sLFOfiltmask,
        genlagtc,
        "glm",
        outputname,
        oversamptr,
        sLFOfitmean,
        rvalue,
        r2value,
        fitNorm[:, :2],
        fitcoeff[:, :2],
        None,
        lagtc,
        None,
        LGR,
        TimingLGR,
        optiondict,
        regressderivs=1,
        timemask=timemask,
        debug=debug,
    )

    medfiltregressderivratios, filteredregressderivratios, delayoffsetMAD = (
        tide_refinedelay.filterderivratios(
            regressderivratios,
            nativespaceshape,
            validvoxels,
            (xdim, ydim, slicethickness),
            gausssigma=gausssigma,
            patchthresh=patchthresh,
            filetype=theinputdata.filetype,
            rt_floattype=rt_floattype,
            debug=debug,
        )
    )

    # find the mapping of derivative ratios to delays
    tide_refinedelay.trainratiotooffset(
        genlagtc,
        initial_fmri_x,
        outputname,
        outputlevel,
        mindelay=mindelay,
        maxdelay=maxdelay,
        numpoints=numpoints,
        debug=debug,
    )

    # now calculate the delay offsets
    delayoffset = np.zeros_like(filteredregressderivratios)
    if debug:
        print(f"calculating delayoffsets for {filteredregressderivratios.shape[0]} voxels")
    for i in range(filteredregressderivratios.shape[0]):
        delayoffset[i], closestoffset = tide_refinedelay.ratiotodelay(
            filteredregressderivratios[i]
        )

    namesuffix = "_desc-delayoffset_hist"
    tide_stats.makeandsavehistogram(
        delayoffset[np.where(sLFOfiltmask > 0)],
        histlen,
        1,
        outputname + namesuffix,
        displaytitle="Histogram of delay offsets calculated from regression coefficients",
        dictvarname="delayoffsethist",
        thedict=optiondict,
    )

    return (
        delayoffset,
        regressderivratios,
        medfiltregressderivratios,
        filteredregressderivratios,
        delayoffsetMAD,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2016-2026 Blaise Frederick
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
import argparse
import io
import tempfile
from unittest.mock import MagicMock, patch

import matplotlib

matplotlib.use("Agg")

import matplotlib.pyplot as plt
import numpy as np
import pytest

from rapidtide.workflows.spectrogram import (
    _get_parser,
    calcspecgram,
    make_legend_axes,
    ndplot,
    showspecgram,
    spectrogram,
)

# ---- helpers ----


def _make_test_signal(samplerate=100.0, duration=5.0, freq=10.0):
    """Create a simple sinusoidal test signal.

    Returns (signal, time_vector).
    """
    npts = int(samplerate * duration)
    t = np.arange(npts) / samplerate
    x = np.sin(2.0 * np.pi * freq * t)
    return x, t


def _make_chirp_signal(samplerate=200.0, duration=5.0, f0=5.0, f1=50.0):
    """Create a chirp signal sweeping from f0 to f1.

    Returns (signal, time_vector).
    """
    npts = int(samplerate * duration)
    t = np.arange(npts) / samplerate
    x = np.sin(2.0 * np.pi * (f0 * t + (f1 - f0) * t**2 / (2.0 * duration)))
    return x, t


def _make_default_args(**overrides):
    """Create default args Namespace for spectrogram workflow."""
    defaults = dict(
        inputfile="input.txt",
        nperseg=128,
        samplerate=100.0,
        debug=False,
    )
    defaults.update(overrides)
    return argparse.Namespace(**defaults)


# ---- _get_parser tests ----


def parser_basic(debug=False):
    """Test that _get_parser returns a valid parser."""
    parser = _get_parser()
    assert isinstance(parser, argparse.ArgumentParser)
    assert parser.prog == "spectrogram"

    if debug:
        print("parser_basic passed")


def parser_required_args(debug=False):
    """Test that parser requires inputfile."""
    parser = _get_parser()
    with pytest.raises(SystemExit):
        parser.parse_args([])

    if debug:
        print("parser_required_args passed")


def parser_defaults(debug=False):
    """Test default values from the parser."""
    parser = _get_parser()
    with tempfile.NamedTemporaryFile(suffix=".txt") as f:
        args = parser.parse_args([f.name])

    assert args.nperseg == 128
    assert args.samplerate is None
    assert args.debug is False

    if debug:
        print("parser_defaults passed")


def parser_nperseg(debug=False):
    """Test --nperseg option."""
    parser = _get_parser()
    with tempfile.NamedTemporaryFile(suffix=".txt") as f:
        args = parser.parse_args([f.name, "--nperseg", "256"])
    assert args.nperseg == 256

    if debug:
        print("parser_nperseg passed")


def parser_samplerate(debug=False):
    """Test --samplerate option."""
    parser = _get_parser()
    with tempfile.NamedTemporaryFile(suffix=".txt") as f:
        args = parser.parse_args([f.name, "--samplerate", "44100.0"])
    assert np.isclose(args.samplerate, 44100.0)

    if debug:
        print("parser_samplerate passed")


def parser_debug(debug=False):
    """Test --debug flag."""
    parser = _get_parser()
    with tempfile.NamedTemporaryFile(suffix=".txt") as f:
        args = parser.parse_args([f.name, "--debug"])
    assert args.debug is True

    if debug:
        print("parser_debug passed")


def parser_inputfile_validation(debug=False):
    """Test that inputfile must exist (is_valid_file check)."""
    parser = _get_parser()
    with pytest.raises(SystemExit):
        parser.parse_args(["/nonexistent/file.txt"])

    if debug:
        print("parser_inputfile_validation passed")


# ---- calcspecgram tests ----


def calcspecgram_basic(debug=False):
    """Test calcspecgram returns correct output types and shapes."""
    x, t = _make_test_signal(samplerate=100.0, duration=5.0, freq=10.0)

    freq, segtimes, thestft, isinvertable = calcspecgram(x, t, nperseg=32)

    assert isinstance(freq, np.ndarray)
    assert isinstance(segtimes, np.ndarray)
    assert isinstance(thestft, np.ndarray)
    assert isinstance(isinvertable, (bool, np.bool_))

    # freq should be 1D
    assert freq.ndim == 1
    # segtimes should be 1D
    assert segtimes.ndim == 1
    # thestft should be 2D (freq x time)
    assert thestft.ndim == 2
    # first dim of stft should match freq length
    assert thestft.shape[0] == len(freq)
    # second dim of stft should match segtimes length
    assert thestft.shape[1] == len(segtimes)

    if debug:
        print(f"calcspecgram_basic: freq shape={freq.shape}, "
              f"segtimes shape={segtimes.shape}, stft shape={thestft.shape}")
        print("calcspecgram_basic passed")


def calcspecgram_frequency_range(debug=False):
    """Test that frequency range is correct for given sample rate."""
    samplerate = 200.0
    x, t = _make_test_signal(samplerate=samplerate, duration=5.0, freq=10.0)

    freq, segtimes, thestft, isinvertable = calcspecgram(x, t, nperseg=64)

    # onesided FFT: max freq should be Nyquist = samplerate/2
    assert freq[0] >= 0.0
    assert freq[-1] <= samplerate / 2.0

    if debug:
        print(f"calcspecgram_frequency_range: freq range [{freq[0]:.2f}, {freq[-1]:.2f}]")
        print("calcspecgram_frequency_range passed")


def calcspecgram_invertible(debug=False):
    """Test that the NOLA condition is satisfied with default settings."""
    x, t = _make_test_signal(samplerate=100.0, duration=5.0, freq=10.0)

    freq, segtimes, thestft, isinvertable = calcspecgram(x, t, nperseg=32)

    assert bool(isinvertable), "Spectrogram with hann window should be invertible"

    if debug:
        print("calcspecgram_invertible passed")


def calcspecgram_nperseg_affects_resolution(debug=False):
    """Test that different nperseg values produce different frequency resolutions."""
    x, t = _make_test_signal(samplerate=100.0, duration=5.0, freq=10.0)

    freq_small, _, stft_small, _ = calcspecgram(x, t, nperseg=16)
    freq_large, _, stft_large, _ = calcspecgram(x, t, nperseg=64)

    # larger nperseg should give more frequency bins
    assert len(freq_large) > len(freq_small)

    if debug:
        print(f"calcspecgram_nperseg: small={len(freq_small)} bins, "
              f"large={len(freq_large)} bins")
        print("calcspecgram_nperseg_affects_resolution passed")


def calcspecgram_stft_is_complex(debug=False):
    """Test that the STFT output is complex."""
    x, t = _make_test_signal(samplerate=100.0, duration=5.0, freq=10.0)

    freq, segtimes, thestft, isinvertable = calcspecgram(x, t, nperseg=32)

    assert np.iscomplexobj(thestft), "STFT output should be complex"

    if debug:
        print("calcspecgram_stft_is_complex passed")


def calcspecgram_peak_at_signal_freq(debug=False):
    """Test that the STFT has a peak near the signal frequency."""
    signal_freq = 10.0
    x, t = _make_test_signal(samplerate=100.0, duration=5.0, freq=signal_freq)

    freq, segtimes, thestft, isinvertable = calcspecgram(x, t, nperseg=64)

    # average magnitude across time to find dominant frequency
    avg_mag = np.mean(np.abs(thestft), axis=1)
    peak_freq = freq[np.argmax(avg_mag)]

    # peak should be near the signal frequency (within frequency resolution)
    freq_resolution = freq[1] - freq[0]
    assert abs(peak_freq - signal_freq) <= freq_resolution, (
        f"Peak at {peak_freq:.2f} Hz, expected near {signal_freq:.2f} Hz"
    )

    if debug:
        print(f"calcspecgram_peak_at_signal_freq: peak at {peak_freq:.2f} Hz")
        print("calcspecgram_peak_at_signal_freq passed")


def calcspecgram_different_window(debug=False):
    """Test calcspecgram with a different window type."""
    x, t = _make_test_signal(samplerate=100.0, duration=5.0, freq=10.0)

    freq, segtimes, thestft, isinvertable = calcspecgram(
        x, t, nperseg=32, windowtype="hamming"
    )

    assert len(freq) > 0
    assert thestft.shape[0] == len(freq)
    assert thestft.shape[1] == len(segtimes)

    if debug:
        print("calcspecgram_different_window passed")


def calcspecgram_short_signal(debug=False):
    """Test calcspecgram with a signal only slightly longer than nperseg."""
    samplerate = 100.0
    npts = 40  # just above nperseg=32
    t = np.arange(npts) / samplerate
    x = np.sin(2.0 * np.pi * 10.0 * t)

    freq, segtimes, thestft, isinvertable = calcspecgram(x, t, nperseg=32)

    assert len(freq) > 0
    assert len(segtimes) > 0
    assert thestft.shape[0] == len(freq)

    if debug:
        print("calcspecgram_short_signal passed")


# ---- make_legend_axes tests ----


def make_legend_axes_basic(debug=False):
    """Test that make_legend_axes returns an axes object."""
    fig, ax = plt.subplots()
    legend_ax = make_legend_axes(ax)

    assert legend_ax is not None
    # the returned object should be a matplotlib axes
    assert hasattr(legend_ax, "set_visible")

    plt.close(fig)

    if debug:
        print("make_legend_axes_basic passed")


def make_legend_axes_separate(debug=False):
    """Test that make_legend_axes creates a distinct axes from the original."""
    fig, ax = plt.subplots()
    legend_ax = make_legend_axes(ax)

    assert legend_ax is not ax

    plt.close(fig)

    if debug:
        print("make_legend_axes_separate passed")


# ---- showspecgram tests ----


def showspecgram_mag(debug=False):
    """Test showspecgram in magnitude mode."""
    x, t = _make_test_signal(samplerate=100.0, duration=2.0, freq=10.0)
    freq, segtimes, thestft, _ = calcspecgram(x, t, nperseg=32)

    fig, ax = plt.subplots()
    im, cbar = showspecgram(thestft, segtimes, freq, ax, fig, mode="mag")

    assert im is not None
    assert cbar is not None

    plt.close(fig)

    if debug:
        print("showspecgram_mag passed")


def showspecgram_phase(debug=False):
    """Test showspecgram in phase mode."""
    x, t = _make_test_signal(samplerate=100.0, duration=2.0, freq=10.0)
    freq, segtimes, thestft, _ = calcspecgram(x, t, nperseg=32)

    fig, ax = plt.subplots()
    im, cbar = showspecgram(thestft, segtimes, freq, ax, fig, mode="phase")

    assert im is not None
    assert cbar is not None

    plt.close(fig)

    if debug:
        print("showspecgram_phase passed")


def showspecgram_real(debug=False):
    """Test showspecgram in real mode."""
    x, t = _make_test_signal(samplerate=100.0, duration=2.0, freq=10.0)
    freq, segtimes, thestft, _ = calcspecgram(x, t, nperseg=32)

    fig, ax = plt.subplots()
    im, cbar = showspecgram(thestft, segtimes, freq, ax, fig, mode="real")

    assert im is not None
    assert cbar is not None

    plt.close(fig)

    if debug:
        print("showspecgram_real passed")


def showspecgram_imag(debug=False):
    """Test showspecgram in imaginary mode."""
    x, t = _make_test_signal(samplerate=100.0, duration=2.0, freq=10.0)
    freq, segtimes, thestft, _ = calcspecgram(x, t, nperseg=32)

    fig, ax = plt.subplots()
    im, cbar = showspecgram(thestft, segtimes, freq, ax, fig, mode="imag")

    assert im is not None
    assert cbar is not None

    plt.close(fig)

    if debug:
        print("showspecgram_imag passed")


def showspecgram_invalid_mode(debug=False):
    """Test showspecgram exits with invalid mode."""
    x, t = _make_test_signal(samplerate=100.0, duration=2.0, freq=10.0)
    freq, segtimes, thestft, _ = calcspecgram(x, t, nperseg=32)

    fig, ax = plt.subplots()
    with pytest.raises(SystemExit):
        showspecgram(thestft, segtimes, freq, ax, fig, mode="invalid")

    plt.close(fig)

    if debug:
        print("showspecgram_invalid_mode passed")


def showspecgram_ylim(debug=False):
    """Test that showspecgram sets ylim from freq[1] to freq.max()."""
    x, t = _make_test_signal(samplerate=100.0, duration=2.0, freq=10.0)
    freq, segtimes, thestft, _ = calcspecgram(x, t, nperseg=32)

    fig, ax = plt.subplots()
    showspecgram(thestft, segtimes, freq, ax, fig, mode="mag")

    ymin, ymax = ax.get_ylim()
    assert np.isclose(ymin, freq[1]), f"ymin should be freq[1]={freq[1]}, got {ymin}"
    assert np.isclose(ymax, freq.max()), f"ymax should be freq.max()={freq.max()}, got {ymax}"

    plt.close(fig)

    if debug:
        print("showspecgram_ylim passed")


def showspecgram_prints_dimensions(debug=False):
    """Test that showspecgram prints time dimension info."""
    x, t = _make_test_signal(samplerate=100.0, duration=2.0, freq=10.0)
    freq, segtimes, thestft, _ = calcspecgram(x, t, nperseg=32)

    fig, ax = plt.subplots()
    captured = io.StringIO()
    with patch("sys.stdout", captured):
        showspecgram(thestft, segtimes, freq, ax, fig, mode="mag")

    output = captured.getvalue()
    # showspecgram prints len(t), len(time)
    assert str(len(segtimes)) in output

    plt.close(fig)

    if debug:
        print("showspecgram_prints_dimensions passed")


# ---- ndplot tests ----


def ndplot_basic(debug=False):
    """Test that ndplot runs without error."""
    x, t = _make_test_signal(samplerate=100.0, duration=2.0, freq=10.0)

    captured = io.StringIO()
    with patch("sys.stdout", captured):
        ndplot(x, t, "Test Signal", nperseg=32)

    output = captured.getvalue()
    assert "arrived in ndplot" in output
    assert "invertable" in output

    plt.close("all")

    if debug:
        print("ndplot_basic passed")


def ndplot_creates_subplots(debug=False):
    """Test that ndplot creates the expected number of subplots."""
    x, t = _make_test_signal(samplerate=100.0, duration=2.0, freq=10.0)

    ndplot(x, t, "Test Signal", nperseg=32)

    # ndplot creates a figure with 3 subplots (311, 312, 313)
    fig = plt.gcf()
    # should have at least 3 axes (plus legend axes)
    assert len(fig.get_axes()) >= 3

    plt.close("all")

    if debug:
        print("ndplot_creates_subplots passed")


def ndplot_different_nperseg(debug=False):
    """Test ndplot with different nperseg values."""
    x, t = _make_test_signal(samplerate=100.0, duration=3.0, freq=10.0)

    # should run without error for various nperseg
    for nperseg in [16, 32, 64]:
        ndplot(x, t, f"nperseg={nperseg}", nperseg=nperseg)
        plt.close("all")

    if debug:
        print("ndplot_different_nperseg passed")


# ---- spectrogram workflow tests ----


def spectrogram_basic(debug=False):
    """Test spectrogram workflow with mocked I/O and parser."""
    x, t = _make_test_signal(samplerate=100.0, duration=3.0, freq=10.0)
    args = _make_default_args(samplerate=100.0, nperseg=32)

    def mock_readvectorsfromtextfile(fname, onecol=True):
        return (100.0, 0.0, None, x, False, "text")

    with (
        patch(
            "rapidtide.workflows.spectrogram._get_parser",
            return_value=MagicMock(
                parse_args=MagicMock(return_value=args),
                print_help=MagicMock(),
            ),
        ),
        patch(
            "rapidtide.workflows.spectrogram.tide_io.readvectorsfromtextfile",
            side_effect=mock_readvectorsfromtextfile,
        ),
        patch("rapidtide.workflows.spectrogram.plt.show"),
    ):
        spectrogram(args)

    plt.close("all")

    if debug:
        print("spectrogram_basic passed")


def spectrogram_samplerate_from_args(debug=False):
    """Test that --samplerate overrides file sample rate."""
    x, t = _make_test_signal(samplerate=100.0, duration=3.0, freq=10.0)
    args = _make_default_args(samplerate=200.0, nperseg=32)

    captured_xvec = {}

    original_ndplot = ndplot

    def mock_ndplot(yvec, xvec, label, nperseg=32):
        captured_xvec["xvec"] = xvec.copy()
        captured_xvec["yvec"] = yvec.copy()

    def mock_readvectorsfromtextfile(fname, onecol=True):
        # file says 100Hz but args say 200Hz
        return (100.0, 0.0, None, x, False, "text")

    with (
        patch(
            "rapidtide.workflows.spectrogram._get_parser",
            return_value=MagicMock(
                parse_args=MagicMock(return_value=args),
                print_help=MagicMock(),
            ),
        ),
        patch(
            "rapidtide.workflows.spectrogram.tide_io.readvectorsfromtextfile",
            side_effect=mock_readvectorsfromtextfile,
        ),
        patch("rapidtide.workflows.spectrogram.ndplot", side_effect=mock_ndplot),
        patch("rapidtide.workflows.spectrogram.plt.show"),
    ):
        spectrogram(args)

    # timestep = 1/200 = 0.005, xvec[1] - xvec[0] should be 0.005
    xvec = captured_xvec["xvec"]
    timestep = xvec[1] - xvec[0]
    assert np.isclose(timestep, 1.0 / 200.0), (
        f"Expected timestep=0.005, got {timestep}"
    )

    if debug:
        print("spectrogram_samplerate_from_args passed")


def spectrogram_samplerate_from_file(debug=False):
    """Test that file sample rate is used when --samplerate is not set."""
    x, _ = _make_test_signal(samplerate=50.0, duration=3.0, freq=10.0)
    args = _make_default_args(samplerate=None, nperseg=32)

    captured_xvec = {}

    def mock_ndplot(yvec, xvec, label, nperseg=32):
        captured_xvec["xvec"] = xvec.copy()

    def mock_readvectorsfromtextfile(fname, onecol=True):
        return (50.0, 0.0, None, x, False, "text")

    with (
        patch(
            "rapidtide.workflows.spectrogram._get_parser",
            return_value=MagicMock(
                parse_args=MagicMock(return_value=args),
                print_help=MagicMock(),
            ),
        ),
        patch(
            "rapidtide.workflows.spectrogram.tide_io.readvectorsfromtextfile",
            side_effect=mock_readvectorsfromtextfile,
        ),
        patch("rapidtide.workflows.spectrogram.ndplot", side_effect=mock_ndplot),
        patch("rapidtide.workflows.spectrogram.plt.show"),
    ):
        spectrogram(args)

    xvec = captured_xvec["xvec"]
    timestep = xvec[1] - xvec[0]
    assert np.isclose(timestep, 1.0 / 50.0), (
        f"Expected timestep=0.02, got {timestep}"
    )

    if debug:
        print("spectrogram_samplerate_from_file passed")


def spectrogram_no_samplerate_exits(debug=False):
    """Test that spectrogram exits when no sample rate is available."""
    x, _ = _make_test_signal(samplerate=100.0, duration=3.0, freq=10.0)
    args = _make_default_args(samplerate=None, nperseg=32)

    def mock_readvectorsfromtextfile(fname, onecol=True):
        return (None, 0.0, None, x, False, "text")

    with pytest.raises(SystemExit):
        with (
            patch(
                "rapidtide.workflows.spectrogram._get_parser",
                return_value=MagicMock(
                    parse_args=MagicMock(return_value=args),
                    print_help=MagicMock(),
                ),
            ),
            patch(
                "rapidtide.workflows.spectrogram.tide_io.readvectorsfromtextfile",
                side_effect=mock_readvectorsfromtextfile,
            ),
            patch("rapidtide.workflows.spectrogram.plt.show"),
        ):
            spectrogram(args)

    if debug:
        print("spectrogram_no_samplerate_exits passed")


def spectrogram_starttime(debug=False):
    """Test that starttime from file offsets the time vector."""
    x, _ = _make_test_signal(samplerate=100.0, duration=2.0, freq=10.0)
    args = _make_default_args(samplerate=100.0, nperseg=32)

    captured_xvec = {}

    def mock_ndplot(yvec, xvec, label, nperseg=32):
        captured_xvec["xvec"] = xvec.copy()

    def mock_readvectorsfromtextfile(fname, onecol=True):
        return (100.0, 5.0, None, x, False, "text")

    with (
        patch(
            "rapidtide.workflows.spectrogram._get_parser",
            return_value=MagicMock(
                parse_args=MagicMock(return_value=args),
                print_help=MagicMock(),
            ),
        ),
        patch(
            "rapidtide.workflows.spectrogram.tide_io.readvectorsfromtextfile",
            side_effect=mock_readvectorsfromtextfile,
        ),
        patch("rapidtide.workflows.spectrogram.ndplot", side_effect=mock_ndplot),
        patch("rapidtide.workflows.spectrogram.plt.show"),
    ):
        spectrogram(args)

    xvec = captured_xvec["xvec"]
    assert np.isclose(xvec[0], 5.0), f"Expected xvec[0]=5.0, got {xvec[0]}"

    if debug:
        print("spectrogram_starttime passed")


def spectrogram_debug_output(debug=False):
    """Test that debug mode prints args."""
    x, _ = _make_test_signal(samplerate=100.0, duration=2.0, freq=10.0)
    args = _make_default_args(samplerate=100.0, nperseg=32, debug=True)

    def mock_readvectorsfromtextfile(fname, onecol=True):
        return (100.0, 0.0, None, x, False, "text")

    captured = io.StringIO()
    with (
        patch(
            "rapidtide.workflows.spectrogram._get_parser",
            return_value=MagicMock(
                parse_args=MagicMock(return_value=args),
                print_help=MagicMock(),
            ),
        ),
        patch(
            "rapidtide.workflows.spectrogram.tide_io.readvectorsfromtextfile",
            side_effect=mock_readvectorsfromtextfile,
        ),
        patch("rapidtide.workflows.spectrogram.plt.show"),
        patch("sys.stdout", captured),
    ):
        spectrogram(args)

    output = captured.getvalue()
    # debug prints args object
    assert "nperseg" in output or "samplerate" in output

    plt.close("all")

    if debug:
        print("spectrogram_debug_output passed")


def spectrogram_uses_inputfile_as_label(debug=False):
    """Test that spectrogram uses inputfile name as the plot label."""
    x, _ = _make_test_signal(samplerate=100.0, duration=2.0, freq=10.0)
    args = _make_default_args(
        inputfile="my_data.txt", samplerate=100.0, nperseg=32,
    )

    captured_label = {}

    def mock_ndplot(yvec, xvec, label, nperseg=32):
        captured_label["label"] = label

    def mock_readvectorsfromtextfile(fname, onecol=True):
        return (100.0, 0.0, None, x, False, "text")

    with (
        patch(
            "rapidtide.workflows.spectrogram._get_parser",
            return_value=MagicMock(
                parse_args=MagicMock(return_value=args),
                print_help=MagicMock(),
            ),
        ),
        patch(
            "rapidtide.workflows.spectrogram.tide_io.readvectorsfromtextfile",
            side_effect=mock_readvectorsfromtextfile,
        ),
        patch("rapidtide.workflows.spectrogram.ndplot", side_effect=mock_ndplot),
        patch("rapidtide.workflows.spectrogram.plt.show"),
    ):
        spectrogram(args)

    assert captured_label["label"] == "my_data.txt"

    if debug:
        print("spectrogram_uses_inputfile_as_label passed")


# ---- main test function ----


def test_spectrogram(debug=False):
    # parser tests
    parser_basic(debug=debug)
    parser_required_args(debug=debug)
    parser_defaults(debug=debug)
    parser_nperseg(debug=debug)
    parser_samplerate(debug=debug)
    parser_debug(debug=debug)
    parser_inputfile_validation(debug=debug)

    # calcspecgram tests
    calcspecgram_basic(debug=debug)
    calcspecgram_frequency_range(debug=debug)
    calcspecgram_invertible(debug=debug)
    calcspecgram_nperseg_affects_resolution(debug=debug)
    calcspecgram_stft_is_complex(debug=debug)
    calcspecgram_peak_at_signal_freq(debug=debug)
    calcspecgram_different_window(debug=debug)
    calcspecgram_short_signal(debug=debug)

    # make_legend_axes tests
    make_legend_axes_basic(debug=debug)
    make_legend_axes_separate(debug=debug)

    # showspecgram tests
    showspecgram_mag(debug=debug)
    showspecgram_phase(debug=debug)
    showspecgram_real(debug=debug)
    showspecgram_imag(debug=debug)
    showspecgram_invalid_mode(debug=debug)
    showspecgram_ylim(debug=debug)
    showspecgram_prints_dimensions(debug=debug)

    # ndplot tests
    ndplot_basic(debug=debug)
    ndplot_creates_subplots(debug=debug)
    ndplot_different_nperseg(debug=debug)

    # spectrogram workflow tests
    spectrogram_basic(debug=debug)
    spectrogram_samplerate_from_args(debug=debug)
    spectrogram_samplerate_from_file(debug=debug)
    spectrogram_no_samplerate_exits(debug=debug)
    spectrogram_starttime(debug=debug)
    spectrogram_debug_output(debug=debug)
    spectrogram_uses_inputfile_as_label(debug=debug)


if __name__ == "__main__":
    test_spectrogram(debug=True)
